import configparser
import glob
from itertools import product
import os
import subprocess

from celery.exceptions import Ignore
from contextlib import nullcontext as does_not_raise
import json
import pytest
from unittest.mock import Mock

from .. import app
from ..tasks import condor
from ..tasks import inference


def test_find_appropriate_cal_env(tmp_path):
    filepath = tmp_path / 'H_CalEnvs.txt'
    t1, cal1 = 1126259462, 'O1_calibration.txt'
    t2, cal2 = 1187008882, 'O2_calibration.txt'
    filepath.write_text('{t1} {cal1}\n{t2} {cal2}'.format(
            t1=t1, cal1=cal1, t2=t2, cal2=cal2))
    dir_name = str(tmp_path)
    trigtime_answer = [
        (t1 - 1, cal1), ((t1 + t2) / 2., cal1), (t2 + 1, cal2)]
    for trigtime, answer in trigtime_answer:
        assert os.path.basename(
            inference._find_appropriate_cal_env(trigtime, dir_name)) == answer


@pytest.mark.parametrize(
    'mc,q,answers',
    [[1., 1.,
      [('analysis', 'engine', 'lalinferencenest'),
       ('analysis', 'nparallel', '4'),
       ('analysis', 'roq', 'True'),
       ('paths', 'roq_b_matrix_directory',
        '/home/cbc/ROQ_data/IMRPhenomPv2/'),
       ('engine', 'approx', 'IMRPhenomPv2pseudoFourPN'),
       ('bayeswave', 'Nchain', '10'),
       ('bayeswave', 'Niter', '100000'),
       ('condor', 'bayeswave_request_memory', '16000'),
       ('condor', 'bayeswavepost_request_memory', '16000'),
       ('bayeswave', 'bw_srate', '4096')]],
     [10., 1.,
      [('condor', 'bayeswave_request_memory', '1000'),
       ('condor', 'bayeswavepost_request_memory', '4000')]],
     [1., 1. / 10.,
      [('analysis', 'engine', 'lalinferencenest'),
       ('analysis', 'nparallel', '4'),
       ('analysis', 'roq', 'True'),
       ('paths', 'roq_b_matrix_directory',
        '/home/cbc/ROQ_data/SEOBNRv4ROQ/'),
       ('engine', 'approx', 'SEOBNRv4_ROMpseudoFourPN')]],
     [1., 1. / 20.,
      [('analysis', 'engine', 'lalinferencemcmc'),
       ('analysis', 'nparallel', '10'),
       ('analysis', 'roq', 'False'),
       ('engine', 'approx', 'SEOBNRv4_ROMpseudoFourPN'),
       ('engine', 'neff', '300')]]]
)
def test_prepare_lalinference_ini(monkeypatch, mc, q, answers):

    def mock_calenv(trigtime, path):
        return 'path_to_calenv'

    monkeypatch.setattr(
        'gwcelery.tasks.inference._find_appropriate_cal_env',
        mock_calenv)
    monkeypatch.setitem(app.conf, 'gracedb_host', 'gracedb.ligo.org')

    m1 = mc * q**(-3. / 5.) * (1 + q)**(1. / 5.)
    m2 = q * m1
    event = {
        'gpstime': 1187008882,
        'graceid': 'G1234',
        'extra_attributes':
            {'SingleInspiral': [{'mass1': m1, 'mass2': m2, 'mchirp': mc}]}
    }
    config = configparser.ConfigParser()
    config.read_string(
        inference.prepare_lalinference_ini(event, 'S1234'))
    for section, key, answer in answers:
        assert config[section][key] == answer


def test_setup_dag_for_lalinference(monkeypatch, tmp_path):
    coinc = b'coinc'
    rundir = str(tmp_path)
    ini = 'ini'
    dag = 'lalinference dag'

    def _subprocess_run(cmd, **kwargs):
        assert cmd[2] == rundir

        path_to_coinc = cmd[4]
        assert os.path.exists(path_to_coinc)
        with open(path_to_coinc, 'rb') as f:
            assert f.read() == coinc

        path_to_ini = cmd[5]
        assert os.path.exists(path_to_ini)
        with open(path_to_ini, 'r') as f:
            assert f.read() == ini

        path_to_psd = cmd[7]
        assert os.path.exists(path_to_psd)
        with open(path_to_psd, 'rb') as f:
            assert f.read() == coinc

        with open(os.path.join(rundir, 'multidag.dag'), 'w') as f:
            f.write(dag)

    upload = Mock()
    monkeypatch.setattr('subprocess.run', _subprocess_run)
    monkeypatch.setattr(
        'gwcelery.tasks.inference.prepare_lalinference_ini',
        Mock(return_value=ini))
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)

    path_to_dag = inference._setup_dag_for_lalinference(
        coinc, rundir, {}, 'S1234')

    assert os.path.exists(path_to_dag)
    with open(path_to_dag, 'r') as f:
        assert f.read() == dag
    upload.assert_called_once()


@pytest.mark.parametrize(
    "host,mode,mc",
    product(
        ['gracedb-playground.ligo.org', 'gracedb.ligo.org'],
        ['production', 'fast_test'],
        [30, 10, 3, 2, 1, 0.1]
    )
)
def test_setup_dag_for_bilby(monkeypatch, tmp_path, host, mode, mc):
    psd = b'psd'
    rundir = str(tmp_path)
    event = {'gpstime': 1187008882, 'graceid': 'G1234',
             'extra_attributes': {'CoincInspiral': {'mchirp': mc}}}
    sid = 'S1234'
    ini = 'bilby configuration ini file'
    dag = 'bilby dag'
    monkeypatch.setitem(app.conf, 'gracedb_host', host)

    def _subprocess_run(cmd, **kwargs):
        if host != 'gracedb.ligo.org':
            assert 'o3replay' in cmd

        assert cmd[2] == os.path.join(
            app.conf['pe_results_path'], sid, 'bilby', mode)

        assert cmd[4] == rundir

        path_to_json = cmd[6]
        assert os.path.exists(path_to_json)
        with open(path_to_json, 'r') as f:
            assert event == json.load(f)

        path_to_psd = cmd[8]
        assert os.path.exists(path_to_psd)
        with open(path_to_psd, 'rb') as f:
            assert f.read() == psd

        if mode == "fast_test":
            assert "FastTest" in cmd

        path_to_settings = cmd[10]
        assert os.path.exists(path_to_settings)
        ans = {
            'summarypages_arguments': {'gracedb': event['graceid'],
                                       'no_ligo_skymap': True},
            'queue': 'Online_PE',
            'accounting_user': 'soichiro.morisaki',
        }
        if host != 'gracedb.ligo.org':
            ans['queue'] = 'Online_PE_MDC'
        else:
            ans['accounting'] = 'ligo.prod.o4.cbc.pe.bilby'
        if mode == 'production':
            ans.update(
                {'sampler_kwargs': {'naccept': 60, 'nlive': 500,
                                    'npool': 24, 'sample': 'acceptance-walk'},
                 'n_parallel': 2,
                 'request_cpus': 24,
                 'spline_calibration_nodes': 10,
                 'request_memory_generation': 8.0}
            )
            if 0.6 < mc < 1.465:
                assert 'lowspin_phenomd_fhigh1024_roq' in cmd
                ans['sampler_kwargs']['naccept'] = 10
            elif mc < 2.243:
                assert 'phenompv2_bns_roq' in cmd
            elif mc < 12:
                assert 'low_q_phenompv2_roq' in cmd
            else:
                assert 'phenomxphm_roq' in cmd
                ans['request_memory_generation'] = 36.0
                ans['request_memory'] = 16.0
        elif mode == 'fast_test' and mc < 3.9:
            ans['request_memory_generation'] = 8.0
        with open(path_to_settings, 'r') as f:
            assert json.load(f) == ans

        with open(os.path.join(rundir, 'bilby_config.ini'), 'w') as f:
            f.write(ini)
        dir = os.path.join(rundir, 'submit')
        os.mkdir(dir)
        with open(os.path.join(dir, 'dag_bilby.submit'), 'w') as f:
            f.write(dag)

    def _upload(filecontents, filename, graceid, message, tags):
        assert filecontents == ini

    upload = Mock(side_effect=_upload)
    monkeypatch.setattr('subprocess.run', _subprocess_run)
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)

    if mc < 0.6:
        with pytest.raises(ValueError):
            inference._setup_dag_for_bilby(psd, rundir, event, sid, mode)
    else:
        path_to_dag = inference._setup_dag_for_bilby(
            psd, rundir, event, sid, mode
        )

        assert os.path.exists(path_to_dag)
        with open(path_to_dag, 'r') as f:
            assert f.read() == dag
        upload.assert_called_once()


def test_setup_dag_for_bilby_unknown_mode(tmp_path):
    with pytest.raises(ValueError):
        inference._setup_dag_for_bilby(
            b'psd', str(tmp_path),
            {'gpstime': 1187008882, 'graceid': 'G1234',
             'extra_attributes': {'CoincInspiral': {'mchirp': 1}}},
            'S1234', 'unknown'
        )


@pytest.mark.parametrize(
    'returncode,expectation',
    [
        (0, does_not_raise()),
        (1, pytest.raises(subprocess.CalledProcessError)),
        (inference._RAPIDPE_NO_GSTLAL_TRIGGER_EXIT_CODE,
         pytest.raises(Ignore)),
    ],
)
def test_setup_dag_for_rapidpe(returncode, expectation, monkeypatch, tmp_path):
    rundir = str(tmp_path)
    dag_filename = 'event_all_iterations.dag'
    dag_content = 'rapidpe dag'

    def _subprocess_run(args, *, capture_output=False, check=False):
        # Simulate failed subprocess
        if returncode != 0:
            raise subprocess.CalledProcessError(
                returncode, args,
                output=b"STDOUT", stderr=b"STDERR",
            )

        path_to_ini = args[1]
        assert os.path.exists(path_to_ini)
        with open(os.path.join(rundir, dag_filename), 'w') as f:
            f.write(dag_content)

    upload = Mock()
    monkeypatch.setattr('subprocess.run', _subprocess_run)
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)

    with expectation:
        path_to_dag = inference._setup_dag_for_rapidpe(rundir, 'S1234')

        if returncode == 0:
            # The DAG should be generated
            with open(path_to_dag, 'r') as f:
                assert f.read() == dag_content
            # The ini file should be uploaded
            upload.assert_called_once()
        else:
            # The ini file and error log should be uploaded
            assert upload.call_count == 2


@pytest.mark.parametrize('pipeline', ['lalinference', 'bilby', 'rapidpe'])
def test_setup_dag_for_failure(monkeypatch, tmp_path, pipeline):
    upload = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)
    monkeypatch.setattr('subprocess.run',
                        Mock(side_effect=subprocess.CalledProcessError(
                            1, ['cmd'], output=b'out', stderr=b'err')))
    monkeypatch.setattr(
        'gwcelery.tasks.inference._find_appropriate_cal_env',
        Mock(return_value='./cal_file.txt'))

    rundir = str(tmp_path)
    event = {
        'gpstime': 1187008882,
        'graceid': 'G1234',
        'extra_attributes': {'SingleInspiral':
                             [{'mass1': 1.4, 'mass2': 1.4, 'mchirp': 1.2}],
                             'CoincInspiral': {'mchirp': 1}}
    }
    with pytest.raises(subprocess.CalledProcessError):
        if pipeline == 'lalinference':
            inference._setup_dag_for_lalinference(
                b'coinc', rundir, event, 'S1234')
        elif pipeline == 'bilby':
            inference._setup_dag_for_bilby(
                (b'psd'), rundir, event, 'S1234', 'production')
        elif pipeline == 'rapidpe':
            inference._setup_dag_for_rapidpe(rundir, 'S1234')
    if pipeline == 'bilby':
        assert upload.call_count == 1
    else:
        # For pipelines except for bilby, an ini file is uploaded before dag
        # generation, and hence the call count is 2.
        assert upload.call_count == 2


@pytest.mark.parametrize(
    'pipeline', ['lalinference', 'bilby', 'rapidpe', 'my_awesome_pipeline'])
def test_dag_prepare_task(monkeypatch, pipeline):
    sid = 'S1234'
    coinc = b'coinc'
    event = {'gpstime': 1187008882, 'graceid': 'G1234'}
    rundir = 'rundir'
    path_to_dag = os.path.join(rundir, 'parameter_estimation.dag')
    kwargs = {'bilby_mode': 'production'}

    def mock_download(filename, gid):
        if filename == 'coinc.xml':
            return coinc

    def _setup_dag_for_lalinference(c, r, e, s):
        assert (c == coinc and r == rundir and e == event and s == sid)
        return path_to_dag

    def _setup_dag_for_bilby(c, r, e, s, m):
        assert c == coinc and r == rundir and e == event and s == sid and \
            m == kwargs['bilby_mode']
        return path_to_dag

    def _setup_dag_for_rapidpe(r, s):
        assert r == rundir and s == sid
        return path_to_dag

    def _subprocess_run(cmd, **kwargs):
        assert cmd == ['condor_submit_dag', '-no_submit', path_to_dag]

    mock_setup_dag_for_lalinference = \
        Mock(side_effect=_setup_dag_for_lalinference)
    mock_setup_dag_for_bilby = Mock(side_effect=_setup_dag_for_bilby)
    mock_setup_dag_for_rapidpe = Mock(side_effect=_setup_dag_for_rapidpe)
    monkeypatch.setattr('gwcelery.tasks.gracedb.download.run', mock_download)
    monkeypatch.setattr('gwcelery.tasks.gracedb.get_event.run',
                        Mock(return_value=event))
    monkeypatch.setattr(
        'gwcelery.tasks.inference._setup_dag_for_lalinference.run',
        mock_setup_dag_for_lalinference)
    monkeypatch.setattr('gwcelery.tasks.inference._setup_dag_for_bilby.run',
                        mock_setup_dag_for_bilby)
    monkeypatch.setattr('gwcelery.tasks.inference._setup_dag_for_rapidpe.run',
                        mock_setup_dag_for_rapidpe)
    monkeypatch.setattr('subprocess.run', _subprocess_run)
    if pipeline in ['lalinference', 'bilby', 'rapidpe']:
        inference.dag_prepare_task(
            rundir, event, sid, pipeline, **kwargs).delay()
        if pipeline == 'lalinference':
            mock_setup_dag_for_lalinference.assert_called_once()
        elif pipeline == 'bilby':
            mock_setup_dag_for_bilby.assert_called_once()
        elif pipeline == 'rapidpe':
            mock_setup_dag_for_rapidpe.assert_called_once()
    else:
        with pytest.raises(NotImplementedError):
            inference.dag_prepare_task(
                rundir, event, sid, pipeline).delay()


@pytest.mark.parametrize(
    'exc,host',
    product(
        [condor.JobAborted(1, 'test'),
         condor.JobFailed(1, 'test'),
         condor.JobRunning({'Cluster': 1234})],
        ['gracedb-playground.ligo.org', 'gracedb.ligo.org']
    )
)
def test_job_error_notification(monkeypatch, tmp_path, exc, host):
    monkeypatch.setitem(app.conf, 'gracedb_host', host)
    filenames = ['pe.log', 'pe.err', 'pe.out']
    for filename in filenames:
        with open(str(tmp_path / filename), 'w') as f:
            f.write('test')
    upload = Mock()
    cleanup = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)
    monkeypatch.setattr(
        'gwcelery.tasks.inference._clean_up_bilby.run', cleanup
    )
    monkeypatch.setattr('subprocess.run', Mock())
    inference.job_error_notification(
        None, exc, 'test', 'S1234', str(tmp_path), 'bilby')
    assert upload.call_count == len(filenames) + 1
    if host != 'gracedb.ligo.org':
        cleanup.assert_called_once()


@pytest.mark.parametrize(
    'pipeline,host',
    product(
        ['lalinference', 'bilby_production', 'bilby_fast_test', 'rapidpe',
         'my_awesome_pipeline'],
        ['gracedb-playground.ligo.org', 'gracedb.ligo.org']
    )
)
def test_dag_finished(monkeypatch, tmp_path, pipeline, host):
    monkeypatch.setitem(app.conf, 'gracedb_host', host)
    sid = 'S1234'
    rundir = str(tmp_path / 'rundir')
    resultdir = str(tmp_path / 'rundir/result')
    sampledir = str(tmp_path / 'rundir/final_result')
    datadir = str(tmp_path / 'rundir/data')
    pe_results_path = str(tmp_path / 'public_html/online_pe')
    monkeypatch.setitem(app.conf, 'pe_results_path', pe_results_path)
    pe_results_path = os.path.join(pe_results_path, sid, pipeline)
    os.makedirs(rundir)
    os.makedirs(resultdir)
    os.makedirs(sampledir)
    os.makedirs(pe_results_path)
    os.makedirs(datadir)

    upload = Mock()
    create_label = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', upload)
    monkeypatch.setattr('gwcelery.tasks.gracedb.create_label.run',
                        create_label)

    if pipeline in [
        'lalinference', 'bilby_production', 'bilby_fast_test', 'rapidpe'
    ]:
        kwargs = {}
        if pipeline == 'lalinference':
            paths = [os.path.join(rundir,
                                  'lalinference_1187008756-1187008882.dag'),
                     os.path.join(rundir, 'glitch_median_PSD_forLI_H1.dat'),
                     os.path.join(rundir, 'glitch_median_PSD_forLI_L1.dat'),
                     os.path.join(rundir, 'glitch_median_PSD_forLI_V1.dat'),
                     os.path.join(rundir, 'posterior_samples.hdf5'),
                     os.path.join(pe_results_path, 'extrinsic.png'),
                     os.path.join(pe_results_path, 'intrinsic.png'),
                     os.path.join(pe_results_path, 'posplots.html')]
        elif pipeline in ['bilby_production', 'bilby_fast_test']:
            kwargs['bilby_mode'] = pipeline[6:]
            pipeline = 'bilby'

            input_sample = os.path.join(sampledir, "test_result.hdf5")
            with open(input_sample, 'wb') as f:
                f.write(b'result')

            samplename = f'Bilby.{kwargs["bilby_mode"]}.posterior_samples.hdf5'
            if kwargs["bilby_mode"] == "production":
                samplename = 'Bilby.posterior_samples.hdf5'
            paths = [os.path.join(sampledir, samplename)]

            def _subprocess_run(cmd):
                assert os.path.exists(cmd[1])
                assert cmd[3] == paths[0]

            monkeypatch.setattr(
                'subprocess.run', Mock(side_effect=_subprocess_run))

            path_to_bilby_config = os.path.join(rundir, "bilby_complete.ini")
            with open(path_to_bilby_config, "w") as f:
                f.write(
                    "spline-calibration-envelope-dict="
                    "{'H1': 'H1_cal.txt', 'L1': 'L1_cal.txt'}\n"
                    "psd-dict={'H1': 'H1_psd.txt', 'L1': 'L1_psd.txt'}\n"
                    "waveform-approximant=IMRPhenomPv2\n"
                    "minimum-frequency=20\n"
                    "reference-frequency=100\n"
                    "webdir=webdir"
                )

            def mock_check_output(args, **kwargs):
                ans = [
                    "summarypages", "--webdir", "webdir/pesummary",
                    "--samples", os.path.join(sampledir, 'test_result.hdf5'),
                    "--gw", "--no_ligo_skymap", "--multi_process", "6",
                    "--config", path_to_bilby_config, "--psd", "H1:H1_psd.txt",
                    "L1:L1_psd.txt", "--calibration", "H1:H1_cal.txt",
                    "L1:L1_cal.txt", "--approximant", "IMRPhenomPv2",
                    "--f_low", "20", "--f_ref", "100", "--label", "online"
                ]
                if host != 'gracedb.ligo.org':
                    ans += ["--disable_interactive", "--disable_expert"]
                else:
                    ans += [
                        "--redshift_method", "exact", "--evolve_spins_forwards"
                    ]
                assert args == ans

            monkeypatch.setattr(
                'gwcelery.tasks.condor.check_output.run', mock_check_output)

            with open(
                os.path.join(datadir, "test_generation_roq_weights.hdf5"),
                "wb"
            ) as f:
                f.write(b'data')
            with open(
                os.path.join(datadir, "test_generation_data_dump.pickle"),
                "wb"
            ) as f:
                f.write(b'data')
        elif pipeline == "rapidpe":
            summary_path = os.path.join(rundir, "summary")
            os.makedirs(summary_path, exist_ok=True)
            paths = [
                os.path.join(summary_path, "p_astro.json"),
                os.path.join(summary_path, "p_astro.png"),
            ]

        else:
            paths = []
        for path in paths:
            with open(path, 'wb') as f:
                f.write(b'result')

        if pipeline == 'rapidpe':
            for event_pipeline in ['gstlal', 'pycbc']:
                kwargs['event_pipeline'] = event_pipeline
                inference.dag_finished(rundir, sid, pipeline, **kwargs)
        else:
            inference.dag_finished(rundir, sid, pipeline, **kwargs)

        if pipeline == "rapidpe":
            # +1 corresponds to summary page link
            assert upload.call_count == 2 * (len(paths) + 1)
        elif pipeline == 'bilby':
            # +1 corresponds to pesummary link
            assert upload.call_count == len(paths) + 1
            n = len(glob.glob(os.path.join(datadir, "*")))
            if host == "gracedb.ligo.org":
                assert n == 2
            else:
                assert n == 0
        else:
            assert upload.call_count == len(paths)

        if pipeline == 'bilby':
            create_label.assert_called_once()
        else:
            create_label.assert_not_called()

    else:
        with pytest.raises(NotImplementedError):
            inference.dag_finished(rundir, sid, pipeline)


@pytest.mark.parametrize(
    'pipeline', ['lalinference', 'bilby', 'rapidpe'])
def test_start_pe(monkeypatch, tmp_path, pipeline):
    path_to_sub = 'pe.dag.condor.sub'

    @app.task
    def mock_task():
        return path_to_sub

    def mock_condor_submit(path):
        assert path == path_to_sub

    if pipeline == 'rapidpe':
        event_pipeline_info = {
            'gstlal': {'sid': 'S1234', 'graceid': 'G1234'},
            'pycbc': {'sid': 'S1235', 'graceid': 'G1235'}}
        for event_pipeline in event_pipeline_info:
            dag_prepare_task = Mock(return_value=mock_task.s())

            condor_submit = Mock(side_effect=mock_condor_submit)
            dag_finished = Mock()
            monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', Mock())
            monkeypatch.setattr('os.path.expanduser', Mock(
                return_value=str(tmp_path)))
            monkeypatch.setattr('gwcelery.tasks.inference.dag_prepare_task',
                                dag_prepare_task)
            monkeypatch.setattr(
                    'gwcelery.tasks.condor.submit.run', condor_submit)
            monkeypatch.setattr(
                    'gwcelery.tasks.inference.dag_finished.run',
                    dag_finished)

            inference.start_pe({
                'graceid': event_pipeline_info[event_pipeline]['graceid'],
                'pipeline': event_pipeline},
                event_pipeline_info[event_pipeline]['sid'], pipeline)
            dag_prepare_task.assert_called_once()
            condor_submit.assert_called_once()
            dag_finished.assert_called_once()

    else:
        dag_prepare_task = Mock(return_value=mock_task.si(),
                                name="dag_prepare_task")

        condor_submit = Mock(side_effect=mock_condor_submit,
                             name="condor_submit")
        dag_finished = Mock(name="dag_finished")
        monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', Mock())
        monkeypatch.setattr('distutils.dir_util.mkpath',
                            Mock(return_value=str(tmp_path)))
        monkeypatch.setattr('gwcelery.tasks.inference.dag_prepare_task',
                            dag_prepare_task)
        monkeypatch.setattr('gwcelery.tasks.condor.submit.run', condor_submit)
        monkeypatch.setattr(
                'gwcelery.tasks.inference.dag_finished.run',
                dag_finished)

        inference.start_pe({'graceid': 'G1234'}, 'S1234', pipeline)
        dag_prepare_task.assert_called_once()
        condor_submit.assert_called_once()
        dag_finished.assert_called_once()


@pytest.mark.parametrize(
    'host', ["gracedb-playground.ligo.org", "gracedb.ligo.org"])
def test_bbh_rate_limit(monkeypatch, tmp_path, host):
    monkeypatch.setitem(app.conf, 'gracedb_host', host)
    monkeypatch.setattr('os.path.expanduser', Mock(return_value=str(tmp_path)))
    event = {'gpstime': 1187008882, 'graceid': 'G1234',
             'extra_attributes': {'CoincInspiral': {'mchirp': 20}}}
    for i in range(6):
        rundir = os.path.join(str(tmp_path), f"{i}/production")
        os.makedirs(rundir)
        os.mkdir(os.path.join(rundir, "data"))
        with open(
            os.path.join(rundir, "data/pe_generation_data_dump.pickle"), "wb"
        ) as f:
            f.write(b"test")
        with open(os.path.join(rundir, "event.json"), "w") as f:
            json.dump(event, f, indent=2)

    path_to_sub = 'pe.dag.condor.sub'

    @app.task
    def mock_task():
        return path_to_sub

    dag_prepare_task = Mock(return_value=mock_task.s())

    def mock_condor_submit(path):
        assert path == path_to_sub

    condor_submit = Mock(side_effect=mock_condor_submit)
    dag_finished = Mock()
    monkeypatch.setattr('gwcelery.tasks.gracedb.upload.run', Mock())
    monkeypatch.setattr('distutils.dir_util.mkpath',
                        Mock(return_value=str(tmp_path)))
    monkeypatch.setattr('gwcelery.tasks.inference.dag_prepare_task',
                        dag_prepare_task)
    monkeypatch.setattr('gwcelery.tasks.condor.submit.run', condor_submit)
    monkeypatch.setattr('gwcelery.tasks.inference.dag_finished.run',
                        dag_finished)

    inference.start_pe(event, 'S1234', 'bilby')
    if host == "gracedb-playground.ligo.org":
        dag_prepare_task.assert_not_called()
        condor_submit.assert_not_called()
        dag_finished.assert_not_called()
    else:
        dag_prepare_task.assert_called_once()
        condor_submit.assert_called_once()
        dag_finished.assert_called_once()


_rapidpe_accounting_prod = 'ligo.prod.o4.cbc.pe.lalinferencerapid'
_rapidpe_accounting_dev = 'ligo.dev.o4.cbc.pe.lalinferencerapid'


@pytest.mark.parametrize(
    'gracedb_host,accounting_group',
    [
        ('gracedb.ligo.org', _rapidpe_accounting_prod),
        ('gracedb-test.ligo.org', _rapidpe_accounting_prod),
        ('gracedb-playground.ligo.org', _rapidpe_accounting_dev),
    ],
)
def test_accounting_group_rapidpe(monkeypatch, tmp_path,
                                  gracedb_host, accounting_group):
    rundir = tmp_path
    event = {}
    superevent_id = 'S1234'
    pe_pipeline = 'rapidpe'
    config_path = os.path.join(rundir, 'rapidpe.ini')

    # Set the GraceDB host appropriately
    monkeypatch.setitem(app.conf, 'gracedb_host', gracedb_host)

    # Confirm the config file doesn't exist yet
    assert not os.path.isfile(config_path)

    # Generates the config file
    inference.dag_prepare_task(rundir, event, superevent_id, pe_pipeline).run()

    # Confirm the config file now exists
    assert os.path.isfile(config_path)

    # Parse the config file to confirm the correct accounting group is set
    config = configparser.ConfigParser()
    config.read(config_path)

    assert config.get('General', 'accounting_group') == accounting_group



from importlib import resources
from lxml import etree
from unittest.mock import patch, call

import pytest

from . import data
from ..tasks import external_triggers
from ..tasks import detchar
from ..util import read_json


@pytest.mark.parametrize('pipeline, path',
                         [['Fermi', 'fermi_grb_gcn.xml'],
                          ['INTEGRAL', 'integral_grb_gcn.xml'],
                          ['INTEGRAL_MDC', 'integral_mdc_gcn.xml'],
                          ['AGILE', 'agile_grb_gcn.xml']])
@patch('gwcelery.tasks.external_skymaps.create_upload_external_skymap.run')
@patch('gwcelery.tasks.external_skymaps.get_upload_external_skymap.run')
@patch('gwcelery.tasks.detchar.dqr_json', return_value='dqrjson')
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.gracedb.create_event.run', return_value={
    'graceid': 'E1', 'gpstime': 1, 'instruments': '', 'pipeline': 'Fermi',
    'search': 'GRB',
    'extra_attributes': {'GRB': {'trigger_duration': 1, 'trigger_id': 123,
                                 'ra': 0., 'dec': 0., 'error_radius': 10.}},
    'links': {'self': 'https://gracedb.ligo.org/events/E356793/'}})
def test_handle_create_grb_event(mock_create_event,
                                 mock_upload, mock_json,
                                 mock_get_upload_external_skymap,
                                 mock_create_upload_external_skymap,
                                 pipeline, path):
    if pipeline == 'INTEGRAL':
        with resources.path(data, path) as p:
            fname = str(p)
        root = etree.parse(fname)
        root.find("./What/Param[@name='TrigID']").attrib['value'] = \
            '123456'.encode()
        text = etree.tostring(root, xml_declaration=True, encoding="UTF-8")
    else:
        text = resources.read_binary(data, path)
    external_triggers.handle_grb_gcn(payload=text)
    mock_create_event.assert_called_once_with(
        filecontents=text,
        search='GRB',
        pipeline='INTEGRAL' if pipeline == 'INTEGRAL_MDC' else pipeline,
        group='External',
        labels=None)
    calls = [
        call(
            '"dqrjson"', 'gwcelerydetcharcheckvectors-E1.json', 'E1',
            'DQR-compatible json generated from check_vectors results'),
        call(
            None, None, 'E1',
            ('Detector state for active instruments is unknown.\n{}'
             'Check looked within -2/+2 seconds of superevent. ').format(
                 detchar.generate_table(
                     'Data quality bits', [], [],
                     ['H1:NO_OMC_DCPD_ADC_OVERFLOW',
                      'H1:NO_DMT-ETMY_ESD_DAC_OVERFLOW',
                      'L1:NO_OMC_DCPD_ADC_OVERFLOW',
                      'L1:NO_DMT-ETMY_ESD_DAC_OVERFLOW',
                      'H1:HOFT_OK', 'H1:OBSERVATION_INTENT',
                      'L1:HOFT_OK', 'L1:OBSERVATION_INTENT',
                      'V1:HOFT_OK', 'V1:OBSERVATION_INTENT',
                      'V1:GOOD_DATA_QUALITY_CAT1'])),
            ['data_quality'])
    ]
    mock_upload.assert_has_calls(calls, any_order=True)
    gcn_type_dict = {'Fermi': 115, 'INTEGRAL': 53, 'INTEGRAL_MDC': 53,
                     'AGILE': 105}
    time_dict = {'Fermi': '2018-05-24T18:35:45',
                 'INTEGRAL': '2017-02-03T19:00:05',
                 'INTEGRAL_MDC': '2023-04-04T06:31:24',
                 'AGILE': '2019-03-19T19:40:49'}
    mock_create_upload_external_skymap.assert_called_once_with(
        {'graceid': 'E1',
         'gpstime': 1,
         'instruments': '',
         'pipeline': 'Fermi',
         'search': 'GRB',
         'extra_attributes': {
             'GRB': {
                 'trigger_duration': 1,
                 'trigger_id': 123,
                 'ra': 0.0,
                 'dec': 0.0,
                 'error_radius': 10.0
                    }
              },
         'links': {
             'self': 'https://gracedb.ligo.org/events/E356793/'
                  }
         },
        gcn_type_dict[pipeline], time_dict[pipeline])


@patch('gwcelery.tasks.gracedb.get_events.run', return_value=[])
@patch('gwcelery.tasks.gracedb.create_event.run', return_value={
    'graceid': 'E1', 'gpstime': 1, 'instruments': '', 'pipeline': 'Fermi',
    'search': 'SubGRB',
    'extra_attributes': {'GRB': {'trigger_duration': None, 'trigger_id': 123,
                                 'ra': 0., 'dec': 0., 'error_radius': 10.}},
    'links': {'self': 'https://gracedb.ligo.org/events/E356793/'}})
@patch('gwcelery.tasks.detchar.check_vectors.run')
@patch('gwcelery.tasks.external_skymaps.get_upload_external_skymap.run')
def test_handle_create_subthreshold_grb_event(mock_get_upload_ext_skymap,
                                              mock_check_vectors,
                                              mock_create_event,
                                              mock_get_events):
    text = resources.read_binary(data, 'fermi_subthresh_grb_lowconfidence.xml')
    external_triggers.handle_grb_gcn(payload=text)
    mock_create_event.assert_not_called()
    text = resources.read_binary(data, 'fermi_subthresh_grb_gcn.xml')
    external_triggers.handle_grb_gcn(payload=text)
    mock_get_events.assert_called_once_with(query=(
                                            'group: External pipeline: '
                                            'Fermi grbevent.trigger_id '
                                            '= "578679123"'))
    # Note that this is the exact ID in the .xml file
    mock_create_event.assert_called_once_with(filecontents=text,
                                              search='SubGRB',
                                              pipeline='Fermi',
                                              group='External',
                                              labels=None)
    mock_check_vectors.assert_called_once()
    mock_get_upload_ext_skymap.assert_called_with(
        {'graceid': 'E1', 'gpstime': 1, 'instruments': '',
         'pipeline': 'Fermi', 'search': 'SubGRB',
         'extra_attributes': {
             'GRB': {'trigger_duration': None, 'trigger_id': 123,
                     'ra': 0., 'dec': 0., 'error_radius': 10.}},
         'links': {'self': 'https://gracedb.ligo.org/events/E356793/'}},
        ('https://gcn.gsfc.nasa.gov/notices_gbm_sub/' +
         'gbm_subthresh_578679393.215999_healpix.fits'))


@pytest.mark.parametrize('filename',
                         ['fermi_noise_gcn.xml',
                          'fermi_noise_gcn_2.xml'])
@patch('gwcelery.tasks.external_skymaps.get_upload_external_skymap.run')
@patch('gwcelery.tasks.gracedb.get_events.run', return_value=[])
@patch('gwcelery.tasks.gracedb.create_event.run', return_value={
    'graceid': 'E1', 'gpstime': 1, 'instruments': '', 'pipeline': 'Fermi',
    'search': 'GRB',
    'extra_attributes': {'GRB': {'trigger_duration': 1, 'trigger_id': 123,
                                 'ra': 0., 'dec': 0., 'error_radius': 10.}},
    'links': {'self': 'https://gracedb.ligo.org/events/E356793/'}})
@patch('gwcelery.tasks.detchar.check_vectors.run')
def test_handle_noise_fermi_event(mock_check_vectors,
                                  mock_create_event,
                                  mock_get_events,
                                  mock_get_upload_external_skymap,
                                  filename):
    text = resources.read_binary(data, filename)
    external_triggers.handle_grb_gcn(payload=text)
    mock_get_events.assert_called_once_with(query=(
                                            'group: External pipeline: '
                                            'Fermi grbevent.trigger_id '
                                            '= "598032876"'))
    # Note that this is the exact ID in the .xml file
    mock_create_event.assert_called_once_with(filecontents=text,
                                              search='GRB',
                                              pipeline='Fermi',
                                              group='External',
                                              labels=['NOT_GRB'])
    mock_check_vectors.assert_called_once()
    mock_get_upload_external_skymap.assert_called_once()


@patch('gwcelery.tasks.external_skymaps.create_external_skymap')
@patch('gwcelery.tasks.external_skymaps.get_upload_external_skymap.run')
@patch('gwcelery.tasks.gracedb.get_events.run', return_value=[])
@patch('gwcelery.tasks.gracedb.create_event.run', return_value={
    'graceid': 'E1', 'gpstime': 1, 'instruments': '', 'pipeline': 'Fermi',
    'search': 'GRB',
    'extra_attributes': {'GRB': {'trigger_duration': 1, 'trigger_id': 123,
                                 'ra': 0., 'dec': 0., 'error_radius': 0.}},
    'links': {'self': 'https://gracedb.ligo.org/events/E356793/'}})
@patch('gwcelery.tasks.detchar.check_vectors.run')
def test_handle_initial_fermi_event(mock_check_vectors,
                                    mock_create_event,
                                    mock_get_events,
                                    mock_get_upload_external_skymap,
                                    mock_create_external_skymap):
    text = resources.read_binary(data, 'fermi_initial_grb_gcn.xml')
    external_triggers.handle_grb_gcn(payload=text)
    mock_get_events.assert_called_once_with(query=(
                                            'group: External pipeline: '
                                            'Fermi grbevent.trigger_id '
                                            '= "548841234"'))
    # Note that this is the exact ID in the .xml file
    mock_create_event.assert_called_once_with(filecontents=text,
                                              search='GRB',
                                              pipeline='Fermi',
                                              group='External',
                                              labels=['NOT_GRB'])
    mock_check_vectors.assert_called_once()
    mock_get_upload_external_skymap.assert_called_once()
    mock_create_external_skymap.assert_not_called()


@pytest.mark.parametrize('filename',
                         ['fermi_grb_gcn.xml',
                          'fermi_noise_gcn.xml',
                          'fermi_subthresh_grb_gcn.xml'])
@patch('gwcelery.tasks.external_skymaps.create_upload_external_skymap.run')
@patch('gwcelery.tasks.external_skymaps.get_upload_external_skymap.run')
@patch('gwcelery.tasks.gracedb.create_label.run')
@patch('gwcelery.tasks.gracedb.remove_label.run')
@patch('gwcelery.tasks.gracedb.replace_event.run')
@patch('gwcelery.tasks.gracedb.get_event.run', return_value=[{
    'graceid': 'E1', 'gpstime': 1, 'instruments': '', 'pipeline': 'Fermi',
    'search': 'GRB',
    'extra_attributes': {'GRB': {'trigger_duration': 1, 'trigger_id': 123,
                                 'ra': 0., 'dec': 0., 'error_radius': 10.}},
    'links': {'self': 'https://gracedb.ligo.org/events/E356793/'}}])
@patch('gwcelery.tasks.gracedb.get_events.run', return_value=[{
    'graceid': 'E1', 'gpstime': 1, 'instruments': '', 'pipeline': 'Fermi',
    'search': 'GRB',
    'extra_attributes': {'GRB': {'trigger_duration': 1, 'trigger_id': 123,
                                 'ra': 0., 'dec': 0., 'error_radius': 15.}},
    'links': {'self': 'https://gracedb.ligo.org/events/E356793/'}}])
def test_handle_replace_grb_event(mock_get_event,
                                  mock_get_events,
                                  mock_replace_event, mock_remove_label,
                                  mock_create_label,
                                  mock_get_upload_external_skymap,
                                  mock_create_upload_external_skymap,
                                  filename):
    text = resources.read_binary(data, filename)
    external_triggers.handle_grb_gcn(payload=text)
    if 'subthresh' in filename:
        mock_replace_event.assert_not_called()
        mock_remove_label.assert_not_called()
        mock_create_label.assert_not_called()
    elif 'grb' in filename:
        mock_replace_event.assert_called_once_with('E1', text)
        mock_remove_label.assert_called_once_with('NOT_GRB', 'E1')
    elif 'noise' in filename:
        mock_replace_event.assert_called_once_with('E1', text)
        mock_create_label.assert_called_once_with('NOT_GRB', 'E1')


@patch('gwcelery.tasks.gracedb.get_group', return_value='CBC')
@patch('gwcelery.tasks.gracedb.create_label.run')
@patch('gwcelery.tasks.gracedb.get_labels',
       return_value=['SKYMAP_READY', 'EM_READY'])
def test_handle_create_skymap_label_from_ext_event(mock_get_labels,
                                                   mock_create_label,
                                                   mock_get_group):
    alert = {"uid": "E1212",
             "alert_type": "label_added",
             "data": {"name": "EM_COINC"},
             "object": {
                 "group": "External",
                 "labels": ["EM_COINC", "EXT_SKYMAP_READY"],
                 "superevent": "S1234"
                       }
             }
    external_triggers.handle_grb_igwn_alert(alert)
    calls = [call('SKYMAP_READY', 'E1212'), call('EM_READY', 'E1212')]
    mock_create_label.assert_has_calls(calls)


@patch('gwcelery.tasks.gracedb.get_group', return_value='CBC')
@patch('gwcelery.tasks.gracedb.create_label.run')
def test_handle_create_skymap_label_from_superevent(mock_create_label,
                                                    mock_get_group):
    alert = {"uid": "S1234",
             "alert_type": "label_added",
             "data": {"name": "SKYMAP_READY"},
             "object": {
                 "group": "CBC",
                 "labels": ["SKYMAP_READY", "EM_READY"],
                 "superevent_id": "S1234",
                 "em_events": ['E1212']
                       }
             }
    external_triggers.handle_grb_igwn_alert(alert)
    calls = [call('SKYMAP_READY', 'E1212'), call('EM_READY', 'E1212')]
    mock_create_label.assert_has_calls(calls)


@pytest.mark.parametrize('pipeline',
                         ['Fermi', 'Swift'])
@patch('gwcelery.tasks.raven.raven_pipeline.run')
@patch('gwcelery.tasks.external_skymaps.create_combined_skymap.run')
@patch('gwcelery.tasks.gracedb.get_superevent.run',
       return_value={
           'superevent_id': 'S1234',
           'preferred_event': 'G1234',
           'preferred_event_data':
               {'group': 'CBC'}
                    })
def test_handle_skymaps_ready(mock_get_superevent,
                              mock_create_combined_skymap,
                              mock_raven_pipeline,
                              pipeline):
    """This test makes sure that once sky maps are available for a coincidence
    that the RAVEN pipeline is rerun to calculate the joint FAR with sky map
    information and check publishing conditions, as well as creating a
    combined sky map.
    """
    alert = {"uid": "E1212",
             "alert_type": "label_added",
             "data": {"name": "EM_READY"},
             "object": {
                 "graceid": "E1212",
                 "group": "External",
                 "labels": ["EM_COINC", "EXT_SKYMAP_READY", "EM_READY"],
                 "superevent": "S1234",
                 "pipeline": pipeline,
                 "search": "GRB"
                       }
             }
    external_triggers.handle_grb_igwn_alert(alert)
    mock_raven_pipeline.assert_called_once_with([{'superevent_id': 'S1234',
                                                  'preferred_event': 'G1234',
                                                  'preferred_event_data':
                                                      {'group': 'CBC'}}],
                                                'E1212', alert['object'],
                                                -5, 1, 'CBC',
                                                use_superevent_skymap=False)
    if pipeline != 'Swift':
        mock_create_combined_skymap.assert_called_once_with(
            'S1234', 'E1212', preferred_event='G1234')
    else:
        mock_create_combined_skymap.assert_not_called()


@patch('gwcelery.tasks.raven.trigger_raven_alert')
@patch('gwcelery.tasks.gracedb.get_superevent',
       return_value={'superevent_id': 'S1234',
                     'preferred_event': 'G1234',
                     'preferred_event_data': {
                         'group': 'CBC'},
                     'time_coinc_far': 1e-9,
                     'space_coinc_far': 1e-10})
def test_handle_label_removed(mock_get_superevent,
                              mock_trigger_raven_alert):
    alert = {"uid": "E1212",
             "alert_type": "label_removed",
             "data": {"name": "NOT_GRB"},
             "object": {
                 "graceid": "E1212",
                 "group": "External",
                 "labels": ["EM_COINC", "EXT_SKYMAP_READY", "SKYMAP_READY"],
                 "superevent": "S1234",
                 "pipeline": "Fermi",
                 "search": "GRB"
                       }
             }
    superevent = {'superevent_id': 'S1234',
                  'preferred_event': 'G1234',
                  'preferred_event_data': {
                      'group': 'CBC'},
                  'time_coinc_far': 1e-9,
                  'space_coinc_far': 1e-10}
    coinc_far_dict = {
                'temporal_coinc_far': 1e-9,
                'spatiotemporal_coinc_far': 1e-10
            }
    external_triggers.handle_grb_igwn_alert(alert)
    mock_trigger_raven_alert.assert_called_once_with(
        coinc_far_dict, superevent, alert['uid'],
        alert['object'], 'CBC'
    )


def _mock_get_event(graceid):
    if graceid == 'E1':
        pipeline = 'Fermi'
        extra_labels = []
    elif graceid == 'E2':
        pipeline = 'Swift'
        extra_labels = []
    elif graceid == 'E3':
        pipeline = 'Fermi'
        extra_labels = ['SKYMAP_READY']
    return {'graceid': graceid,
            'pipeline': pipeline,
            'search': 'GRB',
            'superevent': 'S1',
            'labels': (['EM_COINC', 'EXT_SKYMAP_READY',
                        'EM_READY'] + extra_labels)
            }


def _mock_get_superevent(graceid):
    return {'superevent_id': 'S1',
            'preferred_event': 'G1',
            'preferred_event_data':
                {'group': 'CBC'},
            'em_events': ['E1', 'E2', 'E3'],
            'em_type': 'E1',
            'far': 1e-5,
            'labels': ['COMBINEDSKYMAP_READY', 'RAVEN_ALERT',
                       'EM_COINC', 'GCN_PRELIM_SENT']}


@pytest.mark.parametrize('graceid',
                         ['S1', 'E1', 'E2', 'E3'])
@patch('gwcelery.tasks.raven.raven_pipeline.run')
@patch('gwcelery.tasks.external_skymaps.create_combined_skymap.run')
@patch('gwcelery.tasks.gracedb.get_event', _mock_get_event)
@patch('gwcelery.tasks.gracedb.get_superevent.run', _mock_get_superevent)
def test_handle_rerun_combined_skymap(mock_create_combined_skymap,
                                      mock_raven_pipeline,
                                      graceid):
    """This tests re-creating a combined sky map and rerunning the RAVEN
    pipeline whenever a new GW or external skymap is added after sending the
    first alert (and likely after creating the first combined skymap). For
    superevents (S1), this tests this is run for all related em_events. Note
    the combined sky map is not required for Swift (E2) since these are
    produced in this case."""
    event = (_mock_get_superevent(graceid) if 'S' in graceid else
             _mock_get_event(graceid))
    alert = {
        "data": {"filename": ("skymap.multiorder.fits" if 'S' in graceid
                              else 'fermi_skymap.fits.gz'),
                 "file_version": 1,
                 "comment": ''},
        "uid": graceid,
        "alert_type": "log",
        "object": event
    }
    external_triggers.handle_grb_igwn_alert(alert)
    if graceid == 'S1':
        mock_create_combined_skymap.assert_has_calls(
            [call('S1', 'E1', preferred_event=None),
             call('S1', 'E3', preferred_event=None)]
        )
    elif graceid in {'E1', 'E3'}:
        mock_create_combined_skymap.assert_called_once_with(
            'S1', graceid,
            preferred_event=('G1' if graceid == 'E1' else None))
    else:
        mock_create_combined_skymap.assert_not_called()
    if 'S' in graceid:
        calls = [
            call(
                [_mock_get_event('E1')],
                graceid, event,
                -1, 5, 'CBC',
                use_superevent_skymap=True),
            call(
                [_mock_get_event('E2')],
                graceid, event,
                -1, 5, 'CBC',
                use_superevent_skymap=True),
            call(
                [_mock_get_event('E3')],
                graceid, event,
                -1, 5, 'CBC',
                use_superevent_skymap=True)
        ]
    else:
        calls = [
            call(
                [_mock_get_superevent('S1')],
                graceid, event,
                -5, 1, 'CBC',
                use_superevent_skymap=(graceid == 'E3'))
        ]
    mock_raven_pipeline.assert_has_calls(calls)


@pytest.mark.parametrize(
    "alert_type, comment,"
    "expected_combined_skymap_calls, expected_pipeline_calls,"
    "expected_result",
    [
        ("log",
         "Updated superevent parameters: preferred_event: example", 1, 2,
         True),
        ("log",
         "Some other comment", 0, 0, False),
        ("other_type",
         "Updated superevent parameters: preferred_event: example", 0, 0,
         False),
        ("log",
         "Updated superevent parameters: different_param: example", 0, 0,
         False)
    ]
)
@patch('gwcelery.tasks.raven.raven_pipeline.run')
@patch('gwcelery.tasks.external_skymaps.create_combined_skymap.run')
@patch('gwcelery.tasks.gracedb.get_event', _mock_get_event)
def test_preferred_event_coinc_far_calculation(
        mock_create_combined_skymap,
        mock_raven_pipeline,
        alert_type,
        comment,
        expected_combined_skymap_calls,
        expected_pipeline_calls,
        expected_result):
    alert = {
        "alert_type": alert_type,
        "uid": 'S1',
        "object": {"labels": ["EM_COINC"], "superevent_id": 'S1',
                   "em_events": ["E1", "E2"],
                   "preferred_event": 'G1',
                   "em_type": 'E1',
                   "preferred_event_data": {"group": "CBC"}},
        "data": {"comment": comment, "filename": ""}
    }

    external_triggers.handle_grb_igwn_alert(alert)

    assert (mock_create_combined_skymap.call_count ==
            expected_combined_skymap_calls)
    assert mock_raven_pipeline.call_count == expected_pipeline_calls

    if expected_result:
        mock_create_combined_skymap.assert_called_with(
            'S1', 'E1', preferred_event='G1')
        mock_raven_pipeline.assert_has_calls(
            [call(
                [_mock_get_event('E1')],
                'S1', alert['object'],
                -1, 5, 'CBC',
                use_superevent_skymap=False),
             call(
                 [_mock_get_event('E2')],
                 'S1', alert['object'],
                 -1, 5, 'CBC',
                 use_superevent_skymap=False)])


@pytest.mark.parametrize('labels',
                         [["EM_COINC", "SKYMAP_READY", "RAVEN_ALERT", "ADVOK"],
                          ["EM_COINC", "SKYMAP_READY", "ADVOK"]])
@patch('gwcelery.tasks.gracedb.client')
@patch('gwcelery.tasks.gracedb.get_event.run', return_value={
    'graceid': 'E1234', 'pipeline': 'Fermi'})
@patch('gwcelery.tasks.gracedb.create_label.run')
@patch('gwcelery.tasks.raven.sog_paper_pipeline.run')
def test_handle_sog_manuscript_pipeline(mock_sog_paper_pipeline,
                                        mock_create_label,
                                        mock_get_superevent,
                                        mock_gracedb, labels):
    alert = {"uid": "S1212",
             "alert_type": "label_added",
             "data": {"name": "ADVOK"},
             "object": {
                 "graceid": "S1212",
                 "labels": labels,
                 "superevent_id": "S1212",
                 "space_coinc_far": 1e-9,
                 "em_type": "E1234",
                 "em_events": ["E1234"]}
             }

    # Create IGWN alert
    external_triggers.handle_grb_igwn_alert(alert)

    if 'RAVEN_ALERT' in labels:
        mock_sog_paper_pipeline.assert_called_once_with(
            {'graceid': 'E1234', 'pipeline': 'Fermi'}, alert['object'])
    else:
        mock_sog_paper_pipeline.assert_not_called()


@patch('gwcelery.tasks.detchar.dqr_json', return_value='dqrjson')
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.gracedb.create_event.run', return_value={
    'graceid': 'E1', 'gpstime': 1, 'instruments': '', 'pipeline': 'SNEWS',
    'search': 'Supernova'})
def test_handle_create_snews_event(mock_create_event,
                                   mock_upload, mock_json):
    text = resources.read_binary(data, 'snews_gcn.xml')
    external_triggers.handle_snews_gcn(payload=text)
    mock_create_event.assert_called_once_with(filecontents=text,
                                              search='Supernova',
                                              pipeline='SNEWS',
                                              group='External',
                                              labels=None)
    calls = [
        call(
            '"dqrjson"', 'gwcelerydetcharcheckvectors-E1.json', 'E1',
            'DQR-compatible json generated from check_vectors results'),
        call(
            None, None, 'E1',
            ('Detector state for active instruments is unknown.\n{}'
             'Check looked within -10/+10 seconds of superevent. ').format(
                 detchar.generate_table(
                     'Data quality bits', [], [],
                     ['H1:NO_OMC_DCPD_ADC_OVERFLOW',
                      'H1:NO_DMT-ETMY_ESD_DAC_OVERFLOW',
                      'L1:NO_OMC_DCPD_ADC_OVERFLOW',
                      'L1:NO_DMT-ETMY_ESD_DAC_OVERFLOW',
                      'H1:HOFT_OK', 'H1:OBSERVATION_INTENT',
                      'L1:HOFT_OK', 'L1:OBSERVATION_INTENT',
                      'V1:HOFT_OK', 'V1:OBSERVATION_INTENT',
                      'V1:GOOD_DATA_QUALITY_CAT1'])),
            ['data_quality'])
    ]
    mock_upload.assert_has_calls(calls, any_order=True)


@patch('gwcelery.tasks.gracedb.replace_event.run')
@patch('gwcelery.tasks.gracedb.remove_label.run')
@patch('gwcelery.tasks.gracedb.get_events.run',
       return_value=[{'graceid': 'E1', 'search': 'Supernova', 'gpstime': 100,
                      'pipeline': 'SNEWS'}])
def test_handle_replace_snews_event(mock_get_events, mock_remove_label,
                                    mock_replace_event):
    text = resources.read_binary(data, 'snews_gcn.xml')
    external_triggers.handle_snews_gcn(payload=text)
    mock_replace_event.assert_called_once_with('E1', text)


@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_grb_exttrig_creation(mock_raven_coincidence_search):
    """Test dispatch of an IGWN alert message for an exttrig creation."""
    # Test IGWN alert payload.
    alert = read_json(data, 'igwn_alert_exttrig_creation.json')

    # Run function under test
    external_triggers.handle_grb_igwn_alert(alert)

    # Check that the correct tasks were dispatched.
    mock_raven_coincidence_search.assert_has_calls([
        call('E1234', alert['object'], group='Burst', se_searches=['AllSky']),
        call('E1234', alert['object'], group='CBC', searches=['GRB'])])


@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_subgrb_exttrig_creation(mock_raven_coincidence_search):
    """Test dispatch of an IGWN alert message for an exttrig creation."""
    # Test IGWN alert payload.
    alert = read_json(data, 'igwn_alert_subgrb_creation.json')

    # Run function under test
    external_triggers.handle_grb_igwn_alert(alert)

    # Check that the correct tasks were dispatched.
    mock_raven_coincidence_search.assert_has_calls([
        call('E1234', alert['object'], searches=['SubGRB', 'SubGRBTargeted'],
             se_searches=['AllSky'], pipelines=['Fermi'])])


@patch('gwcelery.tasks.external_skymaps.create_upload_external_skymap')
@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_subgrb_targeted_creation(mock_raven_coincidence_search,
                                         mock_create_upload_external_skymap):
    """Test dispatch of an IGWN alert message for an exttrig creation."""
    # Test IGWN alert payload.
    alert = read_json(data, 'igwn_alert_exttrig_subgrb_targeted_creation.json')

    # Run function under test
    external_triggers.handle_grb_igwn_alert(alert)

    # Check that the correct tasks were dispatched.
    mock_raven_coincidence_search.assert_has_calls([
        call('E1234', alert['object'], se_searches=['AllSky'],
             searches=['SubGRB', 'SubGRBTargeted'],
             pipelines=['Swift'])])


@pytest.mark.parametrize('path',
                         ['igwn_alert_snews_test_creation.json',
                          'igwn_alert_snews_creation.json',
                          'igwn_alert_superevent_creation.json'])
@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_sntrig_creation(mock_raven_coincidence_search, path):
    """Test dispatch of an IGWN alert message for SNEWS alerts
    This now includes both real and test SNEWS events to ensure both are
    ingested correctly, as well as a superevent.
    """
    # Test IGWN alert payload.
    alert = read_json(data, path)

    if 'superevent' in path:
        alert['object']['preferred_event_data']['group'] = 'Burst'

    # Run function under test
    external_triggers.handle_snews_igwn_alert(alert)

    if 'test' in path:
        graceid = 'E1236'
    else:
        graceid = 'E1235'

    if 'superevent' in path:
        mock_raven_coincidence_search.assert_called_once_with(
            'S180616h', alert['object'], group='Burst', searches=['Supernova'],
            pipelines=['SNEWS'])
    elif 'snews' in path:
        mock_raven_coincidence_search.assert_called_once_with(
            graceid, alert['object'], group='Burst', searches=['Supernova'],
            pipelines=['SNEWS'])


@patch('gwcelery.tasks.gracedb.get_superevent',
       return_value={'preferred_event': 'M4634'})
@patch('gwcelery.tasks.gracedb.get_group', return_value='CBC')
@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_superevent_cbc_creation(mock_raven_coincidence_search,
                                        mock_get_group,
                                        mock_get_superevent):
    """Test dispatch of an IGWN alert message for a CBC superevent creation."""
    # Test IGWN alert payload.
    alert = read_json(data, 'igwn_alert_superevent_creation.json')

    # Run function under test
    external_triggers.handle_grb_igwn_alert(alert)

    # Check that the correct tasks were dispatched.
    mock_raven_coincidence_search.assert_has_calls([
        call('S180616h', alert['object'], group='CBC',
             searches=['GRB'], se_searches=[]),
        call('S180616h', alert['object'], group='CBC',
             pipelines=['Fermi'], searches=['SubGRB', 'SubGRBTargeted']),
        call('S180616h', alert['object'], group='CBC',
             pipelines=['Swift'], searches=['SubGRB', 'SubGRBTargeted'])])


@pytest.mark.parametrize('search', ['AllSky', 'BBH', 'IMBH'])
@patch('gwcelery.tasks.gracedb.get_superevent',
       return_value={'preferred_event': 'M4634'})
@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_superevent_burst_creation(mock_raven_coincidence_search,
                                          mock_get_superevent, search):
    """
    Test dispatch of an IGWN alert message for a burst superevent
    creation.
    """
    # Test IGWN alert payload.
    alert = read_json(data, 'igwn_alert_superevent_creation.json')
    alert['object']['preferred_event_data']['group'] = 'Burst'
    alert['object']['preferred_event_data']['search'] = search

    # Run function under test
    external_triggers.handle_grb_igwn_alert(alert)

    if search == 'AllSky':
        # Check that the correct tasks were dispatched.
        mock_raven_coincidence_search.assert_has_calls([
            call('S180616h', alert['object'], group='Burst',
                 searches=['GRB'], se_searches=['AllSky']),
            call('S180616h', alert['object'], group='Burst',
                 pipelines=['Fermi'], searches=['SubGRBTargeted']),
            call('S180616h', alert['object'], group='Burst',
                 pipelines=['Swift'], searches=['SubGRBTargeted'])])
    else:
        mock_raven_coincidence_search.assert_not_called()


def _mock_get_events(query):
    if "12345" in query:
        return [{"graceid": "E12345", "search": "SubGRBTargeted"}]
    else:
        return {}


@pytest.mark.parametrize('alert_type',
                         ['initial', 'update', 'retraction'])
@pytest.mark.parametrize('kafka_path',
                         ['kafka_alert_fermi.json',
                          'kafka_alert_fermi_ignore.json',
                          'kafka_alert_swift.json',
                          'kafka_alert_swift_noloc.json',
                          'kafka_alert_swift_wskymap.json'])
@patch('gwcelery.tasks.gracedb.create_event.run', return_value={
    'graceid': 'E12345', 'gpstime': 100, 'pipeline': 'Fermi',
    'extra_attributes': {'GRB': {'trigger_duration': 1, 'trigger_id': 123,
                                 'ra': 10., 'dec': 20., 'error_radius': 10.}}})
@patch('gwcelery.tasks.external_skymaps.create_upload_external_skymap.run')
@patch('gwcelery.tasks.external_skymaps.get_upload_external_skymap.run')
@patch('gwcelery.tasks.external_skymaps.read_upload_skymap_from_base64.run')
@patch('gwcelery.tasks.gracedb.replace_event.run')
@patch('gwcelery.tasks.gracedb.get_event.run',
       return_value={'graceid': 'E12345'})
@patch('gwcelery.tasks.gracedb.create_label.run')
@patch('gwcelery.tasks.gracedb.get_events.run', side_effect=_mock_get_events)
@patch('gwcelery.tasks.detchar.check_vectors.run')
@patch('gwcelery.tasks.gracedb.upload.run')
@patch('gwcelery.tasks.skymaps.plot_allsky.run')
def test_handle_targeted_kafka_alert(mock_plot_allsky,
                                     mock_gracedb_upload,
                                     mock_check_vectors,
                                     mock_get_events,
                                     mock_create_label,
                                     mock_get_event,
                                     mock_replace_event,
                                     mock_read_upload_skymap_from_base64,
                                     mock_get_upload_external_skymap,
                                     mock_create_upload_external_skymap,
                                     mock_create_event,
                                     alert_type, kafka_path):

    alert = read_json(data, kafka_path)
    alert["alert_type"] = alert_type
    voevent_path = '{}_subgrbtargeted_template.xml'.format(
        'fermi' if 'fermi' in kafka_path else 'swift')
    no_loc = 'noloc' in kafka_path
    replace_event = alert_type in ['update', 'retraction']
    pipeline = 'Fermi' if 'fermi' in kafka_path else 'Swift'
    labels = ['NOT_GRB'] if 'ignore' in kafka_path else None
    # Update VOEvent with sky localization info or ID if needed
    if no_loc or replace_event or labels:
        with resources.path(data, voevent_path) as p:
            fname = str(p)
        root = etree.parse(fname)
        if no_loc:
            # Set template to also be zeros if nothing in kafka alert
            root.find(("./WhereWhen/ObsDataLocation/"
                       "ObservationLocation/AstroCoords/Position2D/Value2/"
                       "C1")).text = str(0.).encode()
            root.find(("./WhereWhen/ObsDataLocation/"
                       "ObservationLocation/AstroCoords/Position2D/Value2/"
                       "C2")).text = str(0.).encode()
            root.find(("./WhereWhen/ObsDataLocation/"
                       "ObservationLocation/AstroCoords/Position2D/"
                       "Error2Radius")).text = str(0.).encode()
        if replace_event:
            # Update ID so we can find a previous event and replace it
            new_id = "12345"
            root.find("./What/Param[@name='TrigID']").attrib['value'] = \
                new_id.encode()
            alert["id"] = [new_id]
        if labels:
            # Overwrite value in GCN template to be that in the kafka packet
            new_far = str(alert['far'])
            root.find("./What/Param[@name='FAR']").attrib['value'] = \
                new_far.encode()
        text = etree.tostring(root, xml_declaration=True, encoding="UTF-8")
        # Add back removed newline
        text += b'\n'
    # If no changes needed, read the file directly
    else:
        text = resources.read_binary(data, voevent_path)

    # Send alert through handler
    external_triggers.handle_targeted_kafka_alert(alert)

    # Check either event was created or updated based on alert type
    if replace_event:
        mock_replace_event.assert_called_once_with("E12345", text)
    else:
        mock_create_event.assert_called_once_with(filecontents=text,
                                                  search='SubGRBTargeted',
                                                  pipeline=pipeline,
                                                  group='External',
                                                  labels=labels)

    # Ensure correct localization method used
    if pipeline == 'Fermi' or 'wskymap' in kafka_path:
        mock_read_upload_skymap_from_base64.assert_called()
    elif 'noloc' in kafka_path:
        mock_read_upload_skymap_from_base64.assert_not_called()
        mock_create_upload_external_skymap.assert_not_called()
    else:
        mock_create_upload_external_skymap.assert_called_once()

    if alert_type == "retraction" or (labels and alert_type == "update"):
        mock_create_label.assert_called_once_with("NOT_GRB", "E12345")
    else:
        mock_create_label.assert_not_called()


@pytest.mark.parametrize('path',
                         ['igwn_alert_superevent_creation.json',
                          'igwn_alert_exttrig_creation.json'])
@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_mdc_creation(mock_raven_coincidence_search,
                             path):
    """Test dispatch of an IGWN alert message for a CBC superevent creation."""
    # Test IGWN alert payload.
    alert = read_json(data, path)
    if 'superevent' in path:
        alert['object']['preferred_event_data']['search'] = 'MDC'
    elif 'exttrig' in path:
        alert['object']['search'] = 'MDC'

    # Run function under test
    external_triggers.handle_grb_igwn_alert(alert)

    # Check that the correct tasks were dispatched.
    if 'superevent' in path:
        calls = [call('S180616h', alert['object'], group='CBC',
                      searches=['MDC'])]
    elif 'exttrig' in path:
        calls = [call('E1234', alert['object'], group='CBC',
                      se_searches=['MDC']),
                 call('E1234', alert['object'], group='Burst',
                      se_searches=['MDC'])]
    mock_raven_coincidence_search.assert_has_calls(calls, any_order=True)


@pytest.mark.parametrize('path',
                         ['igwn_alert_superevent_creation.json',
                          'igwn_alert_snews_creation.json'])
@patch('gwcelery.tasks.raven.coincidence_search')
def test_handle_mdc_sn_creation(mock_raven_coincidence_search,
                                path):
    """Test dispatch of an MDC with the supernovae igwn alert listener,
    both a superevent and external event."""
    # Test IGWN alert payload.
    alert = read_json(data, path)
    if 'superevent' in path:
        alert['object']['preferred_event_data']['search'] = 'MDC'
        alert['object']['preferred_event_data']['group'] = 'Burst'
    elif 'snews' in path:
        alert['object']['search'] = 'MDC'

    # Run function under test
    external_triggers.handle_snews_igwn_alert(alert)

    # Check that the correct tasks were dispatched.
    if 'superevent' in path:
        mock_raven_coincidence_search.assert_called_once_with(
            'S180616h', alert['object'], group='Burst', searches=['MDC'],
            pipelines=['SNEWS'])
    elif 'snews' in path:
        mock_raven_coincidence_search.assert_called_once_with(
            'E1235', alert['object'], group='Burst', se_searches=['MDC'],
            pipelines=['SNEWS'])



"""Application configuration for ``gracedb-test01.ligo.org``.

Inherits all settings from :mod:`gwcelery.conf.playground`, with the exceptions
below.
"""

from . import *  # noqa: F401, F403

igwn_alert_group = 'gracedb-test01'
"""IGWN alert group."""

gracedb_host = 'gracedb-test01.igwn.org'
"""GraceDB host."""

kafka_alert_config = {
    'scimma': {'url': 'kafka://kafka.scimma.org/igwn.gwalert-test01',
               'suffix': 'avro', 'skymap_encoder': lambda _: _}
}
"""Kafka broker configuration details"""

sentry_environment = 'dev'
"""Record this `environment tag
<https://docs.sentry.io/enriching-error-data/environments/>`_ in Sentry log
messages."""

mock_events_simulate_multiple_uploads = True
"""If True, then upload each mock event several times in rapid succession with
random jitter in order to simulate multiple pipeline uploads."""

idq_channels = ['H1:IDQ-FAP_OVL_16_4096',
                'L1:IDQ-FAP_OVL_16_4096']
"""Low-latency iDQ false alarm probability channel names for O3 replay."""



from lxml import etree
from urllib.parse import urlparse
from celery import group
from celery.utils.log import get_logger
from pathlib import Path

from ..import app
from . import alerts
from . import detchar
from . import gcn
from . import gracedb
from . import external_skymaps
from . import igwn_alert
from . import raven

log = get_logger(__name__)


REQUIRED_LABELS_BY_TASK = {
    # EM_READY implies preferred event sky map is available
    'compare': {'EM_READY', 'EXT_SKYMAP_READY', 'EM_COINC'},
    'SoG': {'SKYMAP_READY', 'RAVEN_ALERT', 'ADVOK'}
}
"""These labels should be present on an external event to consider it to
be ready for sky map comparison or for post-alert analsis, such as a
measurment of the speed of gravity (SoG).
"""

FERMI_GRB_CLASS_VALUE = 4
"""This is the index that denote GRBs within Fermi's Flight Position
classification."""

FERMI_GRB_CLASS_THRESH = 50
"""This values denotes the threshold of the most likely Fermi source
classification, above which we will consider a Fermi Flight Position
notice."""


@gcn.handler(gcn.NoticeType.SNEWS,
             queue='exttrig',
             shared=False)
def handle_snews_gcn(payload):
    """Handles the payload from SNEWS alerts.

    Prepares the alert to be sent to graceDB as 'E' events.
    """
    root = etree.fromstring(payload)

    #  Get TrigID and Test Event Boolean
    trig_id = root.find("./What/Param[@name='TrigID']").attrib['value']
    ext_group = 'Test' if root.attrib['role'] == 'test' else 'External'

    event_observatory = 'SNEWS'
    if 'mdc-test_event' in root.attrib['ivorn'].lower():
        search = 'MDC'
    else:
        search = 'Supernova'
    query = 'group: External pipeline: {} grbevent.trigger_id = "{}"'.format(
        event_observatory, trig_id)

    (
        gracedb.get_events.si(query=query)
        |
        _create_replace_external_event_and_skymap.s(
            payload, search, event_observatory, ext_group=ext_group
        )
    ).delay()


@gcn.handler(gcn.NoticeType.FERMI_GBM_ALERT,
             gcn.NoticeType.FERMI_GBM_FLT_POS,
             gcn.NoticeType.FERMI_GBM_GND_POS,
             gcn.NoticeType.FERMI_GBM_FIN_POS,
             gcn.NoticeType.SWIFT_BAT_GRB_POS_ACK,
             gcn.NoticeType.FERMI_GBM_SUBTHRESH,
             gcn.NoticeType.INTEGRAL_WAKEUP,
             gcn.NoticeType.INTEGRAL_REFINED,
             gcn.NoticeType.INTEGRAL_OFFLINE,
             gcn.NoticeType.AGILE_MCAL_ALERT,
             queue='exttrig',
             shared=False)
def handle_grb_gcn(payload):
    """Handles the payload from Fermi, Swift, INTEGRAL, and AGILE MCAL
    GCN notices.

    Filters out candidates likely to be noise. Creates external events
    from the notice if new notice, otherwise updates existing event. Then
    creates and/or grabs external sky map to be uploaded to the external event.

    More info for these notices can be found at:
    Fermi-GBM: https://gcn.gsfc.nasa.gov/fermi_grbs.html
    Fermi-GBM sub: https://gcn.gsfc.nasa.gov/fermi_gbm_subthresh_archive.html
    Swift: https://gcn.gsfc.nasa.gov/swift.html
    INTEGRAL: https://gcn.gsfc.nasa.gov/integral.html
    AGILE-MCAL: https://gcn.gsfc.nasa.gov/agile_mcal.html
    """
    root = etree.fromstring(payload)
    u = urlparse(root.attrib['ivorn'])
    stream_path = u.path

    stream_obsv_dict = {'/SWIFT': 'Swift',
                        '/Fermi': 'Fermi',
                        '/INTEGRAL': 'INTEGRAL',
                        '/AGILE': 'AGILE'}
    event_observatory = stream_obsv_dict[stream_path]

    #  Get TrigID
    if event_observatory == 'INTEGRAL' and \
            not any([x in u.fragment for x in ['O3-replay', 'MDC-test']]):
        #  FIXME: revert all this if INTEGRAL fixes their GCN notices
        #  If INTEGRAL, get trigger ID from ivorn rather than the TrigID field
        #  unless O3 replay or MDC event
        trig_id = u.fragment.split('_')[-1].split('-')[0]
        #  Modify the TrigID field so GraceDB has the correct value
        root.find("./What/Param[@name='TrigID']").attrib['value'] = \
            str(trig_id).encode()
        #  Apply changes to payload delivered to GraceDB
        payload = etree.tostring(root, xml_declaration=True, encoding="UTF-8")
    else:
        try:
            trig_id = \
                root.find("./What/Param[@name='TrigID']").attrib['value']
        except AttributeError:
            trig_id = \
                root.find("./What/Param[@name='Trans_Num']").attrib['value']
    ext_group = 'Test' if root.attrib['role'] == 'test' else 'External'

    notice_type = \
        int(root.find("./What/Param[@name='Packet_Type']").attrib['value'])

    reliability = root.find("./What/Param[@name='Reliability']")
    if reliability is not None and int(reliability.attrib['value']) <= 4:
        return

    #  Check if Fermi trigger is likely noise by checking classification
    #  Most_Likely_Index of 4 is an astrophysical GRB
    #  If not at least 50% chance of GRB we will not consider it for RAVEN
    likely_source = root.find("./What/Param[@name='Most_Likely_Index']")
    likely_prob = root.find("./What/Param[@name='Most_Likely_Prob']")
    not_likely_grb = likely_source is not None and \
        (likely_source.attrib['value'] != FERMI_GRB_CLASS_VALUE
         or likely_prob.attrib['value'] < FERMI_GRB_CLASS_THRESH)

    #  Check if initial Fermi alert. These are generally unreliable and should
    #  never trigger a RAVEN alert, but will give us earlier warning of a
    #  possible coincidence. Later notices could change this.
    initial_gbm_alert = notice_type == gcn.NoticeType.FERMI_GBM_ALERT

    #  Check if Swift has lost lock. If so then veto
    lost_lock = \
        root.find("./What/Group[@name='Solution_Status']" +
                  "/Param[@name='StarTrack_Lost_Lock']")
    swift_veto = lost_lock is not None and lost_lock.attrib['value'] == 'true'

    #  Only send alerts if likely a GRB, is not a low-confidence early Fermi
    #  alert, and if not a Swift veto
    if not_likely_grb or initial_gbm_alert or swift_veto:
        label = 'NOT_GRB'
    else:
        label = None

    ivorn = root.attrib['ivorn']
    if 'subthresh' in ivorn.lower():
        search = 'SubGRB'
    elif 'mdc-test_event' in ivorn.lower():
        search = 'MDC'
    else:
        search = 'GRB'

    if search == 'SubGRB' and event_observatory == 'Fermi':
        skymap_link = \
            root.find("./What/Param[@name='HealPix_URL']").attrib['value']
    else:
        skymap_link = None

    query = 'group: External pipeline: {} grbevent.trigger_id = "{}"'.format(
        event_observatory, trig_id)

    (
        gracedb.get_events.si(query=query)
        |
        _create_replace_external_event_and_skymap.s(
            payload, search, event_observatory,
            ext_group=ext_group, label=label,
            notice_date=root.find("./Who/Date").text,
            notice_type=notice_type,
            skymap_link=skymap_link,
            use_radec=search in {'GRB', 'MDC'}
        )
    ).delay()


@igwn_alert.handler('superevent',
                    'mdc_superevent',
                    'external_fermi',
                    'external_swift',
                    'external_integral',
                    'external_agile',
                    shared=False)
def handle_grb_igwn_alert(alert):
    """Parse an IGWN alert message related to superevents/GRB external triggers
    and dispatch it to other tasks.

    Notes
    -----
    This IGWN alert message handler is triggered by creating a new superevent
    or GRB external trigger event, or a label associated with completeness of
    skymaps:

    * Any new event triggers a coincidence search with
      :meth:`gwcelery.tasks.raven.coincidence_search`.
    * When both a GW and GRB sky map are available during a coincidence,
      indicated by the labels ``EM_READY`` and ``EXT_SKYMAP_READY``
      respectively on the external event, this triggers the spacetime coinc
      FAR to be calculated and a combined GW-GRB sky map is created using
      :meth:`gwcelery.tasks.external_skymaps.create_combined_skymap`.

    """
    # Determine GraceDB ID
    graceid = alert['uid']

    # launch searches
    if alert['alert_type'] == 'new':
        if alert['object'].get('group') == 'External':
            # launch search with MDC events and exit
            if alert['object']['search'] == 'MDC':
                raven.coincidence_search(graceid, alert['object'],
                                         group='CBC', se_searches=['MDC'])
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst', se_searches=['MDC'])
                return

            if alert['object']['search'] in ['SubGRB', 'SubGRBTargeted']:
                # if sub-threshold GRB, launch search with that pipeline
                raven.coincidence_search(
                    graceid, alert['object'],
                    searches=['SubGRB', 'SubGRBTargeted'],
                    se_searches=['AllSky'],
                    pipelines=[alert['object']['pipeline']])
            else:
                # launch standard Burst-GRB search
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst', se_searches=['AllSky'])

                # launch standard CBC-GRB search
                raven.coincidence_search(graceid, alert['object'],
                                         group='CBC', searches=['GRB'])
        elif 'S' in graceid:
            # launch standard GRB search based on group
            gw_group = alert['object']['preferred_event_data']['group']
            search = alert['object']['preferred_event_data']['search']

            # launch search with MDC events and exit
            if alert['object']['preferred_event_data']['search'] == 'MDC':
                raven.coincidence_search(graceid, alert['object'],
                                         group=gw_group, searches=['MDC'])
                return
            # Don't run search for BBH or IMBH Burst search
            elif gw_group == 'Burst' and search.lower() != 'allsky':
                return

            se_searches = [] if gw_group == 'CBC' else ['AllSky']
            searches = (['SubGRB', 'SubGRBTargeted'] if gw_group == 'CBC' else
                        ['SubGRBTargeted'])
            # launch standard GRB search
            raven.coincidence_search(graceid, alert['object'],
                                     group=gw_group, searches=['GRB'],
                                     se_searches=se_searches)
            # launch subthreshold search for Fermi and Swift separately to use
            # different time windows, for both CBC and Burst
            for pipeline in ['Fermi', 'Swift']:
                raven.coincidence_search(
                    graceid, alert['object'], group=gw_group,
                    searches=searches, pipelines=[pipeline])
    # re-run raven pipeline and create combined sky map (if not a Swift event)
    # when sky maps are available
    elif alert['alert_type'] == 'label_added' and \
            alert['object'].get('group') == 'External':
        if _skymaps_are_ready(alert['object'], alert['data']['name'],
                              'compare'):
            # if both sky maps present and a coincidence, re-run RAVEN
            # pipeline and create combined sky maps
            ext_event = alert['object']
            superevent_id, ext_id = _get_superevent_ext_ids(
                graceid, ext_event)
            superevent = gracedb.get_superevent(superevent_id)
            _relaunch_raven_pipeline_with_skymaps(
                superevent, ext_event, graceid)
        elif 'EM_COINC' in alert['object']['labels']:
            # if not complete, check if GW sky map; apply label to external
            # event if GW sky map
            se_labels = gracedb.get_labels(alert['object']['superevent'])
            if 'SKYMAP_READY' in se_labels:
                gracedb.create_label.si('SKYMAP_READY', graceid).delay()
            if 'EM_READY' in se_labels:
                gracedb.create_label.si('EM_READY', graceid).delay()
    # apply labels from superevent to external event to update state
    # and trigger functionality requiring sky maps, etc.
    elif alert['alert_type'] == 'label_added' and 'S' in graceid:
        if 'SKYMAP_READY' in alert['object']['labels']:
            # if sky map in superevent, apply label to all external events
            # at the time
            group(
                gracedb.create_label.si('SKYMAP_READY', ext_id)
                for ext_id in alert['object']['em_events']
            ).delay()
        if 'EM_READY' in alert['object']['labels']:
            # if sky map not in superevent but in preferred event, apply label
            # to all external events at the time
            group(
                gracedb.create_label.si('EM_READY', ext_id)
                for ext_id in alert['object']['em_events']
            ).delay()
        if _skymaps_are_ready(alert['object'], alert['data']['name'], 'SoG') \
                and alert['object']['space_coinc_far'] is not None:
            # if a superevent is vetted by ADVOK and a spatial joint FAR is
            # available, check if SoG publishing conditions are met
            (
                gracedb.get_event.si(alert['object']['em_type'])
                |
                raven.sog_paper_pipeline.s(alert['object'])
            ).delay()
    # if new GW or external sky map after first being available, try to remake
    # combine sky map and rerun raven pipeline
    elif alert['alert_type'] == 'log' and \
            'EM_COINC' in alert['object']['labels'] and \
            'fit' in alert['data']['filename'] and \
            'flat' not in alert['data']['comment'].lower() and \
            (alert['data']['filename'] not in
             {external_skymaps.COMBINED_SKYMAP_FILENAME_MULTIORDER,
              external_skymaps.COMBINED_SKYMAP_FILENAME_FLAT}):
        superevent_id, external_id = _get_superevent_ext_ids(
                                         graceid, alert['object'])
        if 'S' in graceid:
            superevent = alert['object']
        else:
            superevent = gracedb.get_superevent(alert['object']['superevent'])
            external_event = alert['object']
        # check if combined sky map already made, with the exception of Swift
        # which will fail
        if 'S' in graceid:
            # Rerun for all eligible external events
            for ext_id in superevent['em_events']:
                external_event = gracedb.get_event(ext_id)
                if REQUIRED_LABELS_BY_TASK['compare'].issubset(
                        set(external_event['labels'])):
                    _relaunch_raven_pipeline_with_skymaps(
                        superevent, external_event, graceid,
                        use_superevent_skymap=True)
        else:
            if REQUIRED_LABELS_BY_TASK['compare'].issubset(
                    set(external_event['labels'])):
                _relaunch_raven_pipeline_with_skymaps(
                    superevent, external_event, graceid)
    # Rerun the coincidence FAR calculation if possible with combined sky map
    # if the preferred event changes
    # We don't want to run this logic if PE results are present
    elif alert['alert_type'] == 'log' and \
            'PE_READY' not in alert['object']['labels'] and \
            'EM_COINC' in alert['object']['labels']:
        new_log_comment = alert['data'].get('comment', '')
        if 'S' in graceid and \
            new_log_comment.startswith('Updated superevent parameters: '
                                       'preferred_event: '):
            superevent = alert['object']
            # Rerun for all eligible external events
            for ext_id in superevent['em_events']:
                external_event = gracedb.get_event(ext_id)
                if REQUIRED_LABELS_BY_TASK['compare'].issubset(
                        set(external_event['labels'])):
                    _relaunch_raven_pipeline_with_skymaps(
                        superevent, external_event, graceid,
                        use_superevent_skymap=False)
    elif alert['alert_type'] == 'label_removed' and \
            alert['object'].get('group') == 'External':
        if alert['data']['name'] == 'NOT_GRB' and \
                'EM_COINC' in alert['object']['labels']:
            # if NOT_GRB is removed, re-check publishing conditions
            superevent_id = alert['object']['superevent']
            superevent = gracedb.get_superevent(superevent_id)
            gw_group = superevent['preferred_event_data']['group']
            coinc_far_dict = {
                'temporal_coinc_far': superevent['time_coinc_far'],
                'spatiotemporal_coinc_far': superevent['space_coinc_far']
            }
            raven.trigger_raven_alert(coinc_far_dict, superevent, graceid,
                                      alert['object'], gw_group)


@igwn_alert.handler('superevent',
                    'mdc_superevent',
                    'external_snews',
                    shared=False)
def handle_snews_igwn_alert(alert):
    """Parse an IGWN alert message related to superevents/SN external triggers
    and dispatch it to other tasks.

    Notes
    -----
    This igwn_alert message handler is triggered by creating a new superevent
    or SN external trigger event:

    * Any new event triggers a coincidence search with
      :meth:`gwcelery.tasks.raven.coincidence_search`.

    """
    # Determine GraceDB ID
    graceid = alert['uid']

    if alert['alert_type'] == 'new':
        if alert['object'].get('superevent_id'):
            group = alert['object']['preferred_event_data']['group']
            search = alert['object']['preferred_event_data']['search']
            searches = ['MDC'] if search == 'MDC' else ['Supernova']
            # Run only on Test and Burst superevents
            if group in {'Burst', 'Test'}:
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst', searches=searches,
                                         pipelines=['SNEWS'])
        else:
            # Run on SNEWS event, either real or test
            search = alert['object']['search']
            if search == 'MDC':
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst',
                                         se_searches=['MDC'],
                                         pipelines=['SNEWS'])
            elif search == 'Supernova':
                raven.coincidence_search(graceid, alert['object'],
                                         group='Burst', searches=['Supernova'],
                                         pipelines=['SNEWS'])


@alerts.handler('fermi',
                'swift')
def handle_targeted_kafka_alert(alert):
    """Parse an alert sent via Kafka from a MOU partner in our joint
    subthreshold targeted search.

    Parameters
    ----------
    alert : dict
        Kafka alert packet

    """
    # Convert alert to VOEvent format
    # FIXME: This is required until native ingesting of kafka events in GraceDB
    payload, pipeline, time, trig_id = _kafka_to_voevent(alert)

    # Veto events that don't pass GRB FAR threshold
    far_grb = alert['far']
    veto_event = \
        app.conf['raven_targeted_far_thresholds']['GRB'][pipeline] < far_grb
    label = ('NOT_GRB' if alert['alert_type'] == "retraction" or veto_event
             else None)

    # Look whether a previous event with the same ID exists
    query = 'group: External pipeline: {} grbevent.trigger_id = "{}"'.format(
        pipeline, trig_id)

    (
        gracedb.get_events.si(query=query)
        |
        _create_replace_external_event_and_skymap.s(
            payload, 'SubGRBTargeted', pipeline,
            label=label, notice_date=time,
            skymap=alert.get('healpix_file'),
            use_radec=('ra' in alert and 'dec' in alert)
        )
    ).delay()


def _skymaps_are_ready(event, label, task):
    label_set = set(event['labels'])
    required_labels = REQUIRED_LABELS_BY_TASK[task]
    return required_labels.issubset(label_set) and label in required_labels


def _get_superevent_ext_ids(graceid, event):
    if 'S' in graceid:
        se_id = event['superevent_id']
        ext_id = event['em_type']
    else:
        se_id = event['superevent']
        ext_id = event['graceid']
    return se_id, ext_id


@app.task(shared=False)
def _launch_external_detchar(event):
    start = event['gpstime']
    if event['pipeline'] == 'SNEWS':
        start, end = event['gpstime'], event['gpstime']
    else:
        integration_time = \
            event['extra_attributes']['GRB']['trigger_duration'] or 4.0
        end = start + integration_time
    detchar.check_vectors.si(event, event['graceid'], start, end).delay()

    return event


def _relaunch_raven_pipeline_with_skymaps(superevent, ext_event, graceid,
                                          use_superevent_skymap=None):
    """Relaunch the RAVEN sky map comparison workflow, include recalculating
    the joint FAR with updated sky map info and create a new combined sky map.

    Parameters
    ----------
    superevent: dict
        superevent dictionary
    exttrig: dict
        external event dictionary
    graceid: str
        GraceDB ID of event
    use_superevent_skymap: bool
        If True/False, use/don't use skymap info from superevent.
        Else if None, check SKYMAP_READY label in external event.

    """
    gw_group = superevent['preferred_event_data']['group']
    tl, th = raven._time_window(graceid, gw_group,
                                [ext_event['pipeline']],
                                [ext_event['search']])
    # FIXME: both overlap integral and combined sky map could be
    # done by the same function since they are so similar
    use_superevent = (use_superevent_skymap
                      if use_superevent_skymap is not None else
                      'SKYMAP_READY' in ext_event['labels'])
    canvas = raven.raven_pipeline.si(
                 [ext_event] if 'S' in graceid else [superevent],
                 graceid,
                 superevent if 'S' in graceid else ext_event,
                 tl, th, gw_group, use_superevent_skymap=use_superevent)
    # Swift localizations are incredibly well localized and require
    # a different method from Fermi/Integral/AGILE
    # FIXME: Add Swift localization information in the future
    if ext_event['pipeline'] != 'Swift':
        # Create new updated combined sky map
        canvas |= external_skymaps.create_combined_skymap.si(
                      superevent['superevent_id'], ext_event['graceid'],
                      preferred_event=(
                          None if use_superevent
                          else superevent['preferred_event']))
    canvas.delay()


@app.task(shared=False)
def _create_replace_external_event_and_skymap(
        events, payload, search, pipeline,
        label=None, ext_group='External', notice_date=None, notice_type=None,
        skymap=None, skymap_link=None, use_radec=False):
    """Either create a new external event or replace an old one if applicable
    Then either uploads a given sky map, try to download one given a link, or
    create one given coordinates.

    Parameters
    ----------
    events : list
        List of external events sharing the same trigger ID
    payload : str
        VOEvent of event being considered
    search : str
        Search of external event
    pipeline : str
        Pipeline of external evevent
    label : list
        Label to be uploaded along with external event. If None, removes
        'NOT_GRB' label from event
    ext_group : str
        Group of external event, 'External' or 'Test'
    notice_date : str
        Time and date of external event
    notice_type : str
        GCN Notice type of external event
    skymap : str
        Base64 encoded sky map
    skymap_link : str
        Link to external sky map to be downloaded
    use_radec : bool
        If true, try to create sky map using given coordinates

    """
    skymap_detchar_canvas = ()
    # If previous event, try to append
    if events and ext_group == 'External':
        assert len(events) == 1, 'Found more than one matching GraceDB entry'
        event, = events
        graceid = event['graceid']
        if label:
            create_replace_canvas = gracedb.create_label.si(label, graceid)
        else:
            create_replace_canvas = gracedb.remove_label.si('NOT_GRB', graceid)

        # Prevent SubGRBs from appending GRBs, also append if same search
        if search == 'GRB' or search == event['search']:
            # Replace event and pass already existing event dictionary
            create_replace_canvas |= gracedb.replace_event.si(graceid, payload)
            create_replace_canvas |= gracedb.get_event.si(graceid)
        else:
            # If not appending just exit
            return

    # If new event, create new entry in GraceDB and launch detchar
    else:
        create_replace_canvas = gracedb.create_event.si(
            filecontents=payload,
            search=search,
            group=ext_group,
            pipeline=pipeline,
            labels=[label] if label else None)
        skymap_detchar_canvas += _launch_external_detchar.s(),

    # Use sky map if provided
    if skymap:
        skymap_detchar_canvas += \
            external_skymaps.read_upload_skymap_from_base64.s(skymap),
    else:
        # Otherwise grab sky map from provided link
        if skymap_link:
            skymap_detchar_canvas += \
                external_skymaps.get_upload_external_skymap.s(skymap_link),
        # Otherwise if threshold Fermi try to grab sky map
        elif pipeline == 'Fermi' and search == 'GRB':
            skymap_detchar_canvas += \
                external_skymaps.get_upload_external_skymap.s(None),
        # Otherwise create sky map from given coordinates
        if use_radec:
            skymap_detchar_canvas += \
                external_skymaps.create_upload_external_skymap.s(
                    notice_type, notice_date),

    (
        create_replace_canvas
        |
        group(skymap_detchar_canvas)
    ).delay()


def _kafka_to_voevent(alert):
    # Define basic values
    pipeline = alert['mission']
    trigger_time = alert['trigger_time']
    alert_time = alert['alert_datetime']
    far = alert['far']
    integration_time = alert['rate_duration']
    id = '_'.join(str(x) for x in alert['id'])

    # sky localization may not be available
    ra = alert.get('ra')
    dec = alert.get('dec')
    # Try to get dec first then ra, None if both misssing
    error = alert.get('dec_uncertainty')
    if error is None:
        error = alert.get('ra_uncertainty')
    # Argument should be list if not None
    if isinstance(error, list):
        error = error[0]
    # if any missing sky map info, set to zeros so will be ignored later
    if ra is None or dec is None or error is None:
        ra, dec, error = 0., 0., 0.

    # Load template
    fname = str(Path(__file__).parent /
                '../tests/data/{}_subgrbtargeted_template.xml'.format(
                    pipeline.lower()))
    root = etree.parse(fname)

    # Update template values
    # Change ivorn to indicate this is a subthreshold targeted event
    root.xpath('.')[0].attrib['ivorn'] = \
        'ivo://lvk.internal/{0}#targeted_subthreshold-{1}'.format(
            pipeline.lower(), trigger_time).encode()

    # Update ID
    root.find("./What/Param[@name='TrigID']").attrib['value'] = \
        id.encode()

    # Change times to chosen time
    root.find("./Who/Date").text = str(alert_time).encode()
    root.find(("./WhereWhen/ObsDataLocation/"
               "ObservationLocation/AstroCoords/Time/TimeInstant/"
               "ISOTime")).text = str(trigger_time).encode()

    root.find("./What/Param[@name='FAR']").attrib['value'] = \
        str(far).encode()

    root.find("./What/Param[@name='Integ_Time']").attrib['value'] = \
        str(integration_time).encode()

    # Sky position
    root.find(("./WhereWhen/ObsDataLocation/"
               "ObservationLocation/AstroCoords/Position2D/Value2/"
               "C1")).text = str(ra).encode()
    root.find(("./WhereWhen/ObsDataLocation/"
               "ObservationLocation/AstroCoords/Position2D/Value2/"
               "C2")).text = str(dec).encode()
    root.find(("./WhereWhen/ObsDataLocation/"
               "ObservationLocation/AstroCoords/Position2D/"
               "Error2Radius")).text = str(error).encode()

    return (etree.tostring(root, xml_declaration=True, encoding="UTF-8",
                           pretty_print=True),
            pipeline, trigger_time.replace('Z', ''), id)



"""Create mock external events to be in coincidence
   with MDC superevents."""

from astropy.time import Time
from lxml import etree
import numpy as np
from pathlib import Path
import random
import re

from ..import app
from . import external_triggers
from . import igwn_alert
from . import raven
from ..tests import data
from ..util import read_json


def create_upload_external_event(gpstime, pipeline, ext_search):
    """ Create a random external event VOEvent.

    Parameters
    ----------
    gpstime : float
        Event's gps time
    pipeline : str
        External trigger pipeline name
    ext_search : str
        Search field for external event
    """
    new_date = str(Time(gpstime, format='gps', scale='utc').isot) + 'Z'
    new_TrigID = str(int(gpstime))
    ra = random.uniform(0, 360)
    thetas = np.arange(-np.pi / 2, np.pi / 2, .01)
    dec = random.choices(np.rad2deg(thetas),
                         weights=np.cos(thetas) / sum(np.cos(thetas)))[0]
    error = .05 if pipeline == 'Swift' else random.uniform(1, 30)

    if pipeline in {'Fermi', 'Swift', 'INTEGRAL', 'AGILE'}:
        is_grb = True
    else:
        is_grb = False

    # If SubGRBTargeted modify alert packet from Kafka
    if ext_search == 'SubGRBTargeted':
        if pipeline == 'Fermi':
            # Template based on:
            # https://github.com/joshuarwood/gcn-schema/tree/main/gcn/notices/
            # fermi/gbm
            alert = read_json(data, 'kafka_alert_fermi.json')
            # Remove sky map file to create our own sky map
            alert.pop('healpix_file')
            alert['dec_uncertainty'] = [error]
        elif pipeline == 'Swift':
            # Template based on:
            # https://github.com/nasa-gcn/gcn-schema/tree/main/gcn/notices/
            # swift/bat/guano
            alert = read_json(data, 'kafka_alert_swift.json')
            alert['ra_uncertainty'] = [error]
        else:
            raise ValueError(
                'Only use Fermi or Swift for SubGRBTargeted search')

        alert['trigger_time'] = new_date
        alert['alert_datetime'] = new_date
        alert['id'] = [new_TrigID]
        alert['ra'] = ra
        alert['dec'] = dec
        # Generate FAR from max threshold and trials factors
        alert['far'] = \
            (app.conf['raven_targeted_far_thresholds']['GRB'][pipeline] *
             random.uniform(0, 1))

        external_triggers.handle_targeted_kafka_alert(alert)

        # Return VOEvent for testing until GraceDB natively ingests kafka
        # alerts
        return external_triggers._kafka_to_voevent(alert)[0]

    # Otherwise modify respective VOEvent template
    else:
        fname = str(Path(__file__).parent /
                    '../tests/data/{0}{1}_{2}gcn.xml'.format(
                        pipeline.lower(),
                        '_subthresh' if ext_search == 'SubGRB' else '',
                        'grb_' if is_grb else ''))

        root = etree.parse(fname)
        # Change ivorn to indicate if this is an MDC event or O3 replay event
        root.xpath('.')[0].attrib['ivorn'] = \
            'ivo://lvk.internal/{0}#{1}{2}_event_{3}'.format(
                pipeline if pipeline != 'Swift' else 'SWIFT',
                '_subthresh' if ext_search == 'SubGRB' else '',
                'MDC-test' if ext_search == 'MDC' else 'O3-replay',
                new_date).encode()

        # Change times to chosen time
        root.find("./Who/Date").text = str(new_date).encode()
        root.find(("./WhereWhen/ObsDataLocation/"
                   "ObservationLocation/AstroCoords/Time/TimeInstant/"
                   "ISOTime")).text = str(new_date).encode()
        if ext_search == 'SubGRB':
            root.find("./What/Param[@name='Trans_Num']").attrib['value'] = \
                str(new_TrigID).encode()
        else:
            root.find("./What/Param[@name='TrigID']").attrib['value'] = \
                str(new_TrigID).encode()

        if is_grb:
            # Give random sky position
            root.find(("./WhereWhen/ObsDataLocation/"
                       "ObservationLocation/AstroCoords/Position2D/Value2/"
                       "C1")).text = str(ra).encode()
            root.find(("./WhereWhen/ObsDataLocation/"
                       "ObservationLocation/AstroCoords/Position2D/Value2/"
                       "C2")).text = str(dec).encode()
            if pipeline != 'Swift':
                root.find(
                    ("./WhereWhen/ObsDataLocation/"
                     "ObservationLocation/AstroCoords/Position2D/"
                     "Error2Radius")).text = str(error).encode()

        event = etree.tostring(root, xml_declaration=True, encoding="UTF-8",
                               pretty_print=True)

        # Upload as from GCN
        if is_grb:
            external_triggers.handle_grb_gcn(event)
        else:
            external_triggers.handle_snews_gcn(event)

        return event


def _offset_time(gpstime, group, pipeline, ext_search):
    """ This function checks coincident time windows for superevents if they
    are of Burst or CBC group.

       Parameters
       ----------
       gpstime : float
           Event's gps time
       group : str
           Burst or CBC
    """
    tl, th = raven._time_window('S1', group, [pipeline], [ext_search])
    return gpstime + random.uniform(tl, th)


def _is_joint_mdc(graceid, se_search):
    """Upload external events to the user-defined frequency of MDC or AllSky
    superevents.

    Looks at the ending letters of a superevent (e.g. 'ac' from 'MS190124ac'),
    converts to a number, and checks if divisible by a number given in the
    configuration file.

    For example, if the configuration number
    :obj:`~gwcelery.conf.joint_mdc_freq` is 10,
    this means joint events with superevents ending with 'j', 't', 'ad', etc.
    """
    end_string = re.split(r'\d+', graceid)[-1].lower()
    val = 0
    for i in range(len(end_string)):
        val += (ord(end_string[i]) - 96) * 26 ** (len(end_string) - i - 1)
    return val % int(app.conf['joint_mdc_freq']) == 0 if se_search == 'MDC' \
        else val % int(app.conf['joint_O3_replay_freq']) == 0


@igwn_alert.handler('mdc_superevent',
                    'superevent',
                    shared=False)
def upload_external_event(alert, ext_search=None):
    """Upload a random GRB event for a certain percentage of MDC
    or O3-replay superevents.

    Notes
    -----
    Every n superevents, upload a GRB candidate within the
    standard CBC-GRB or Burst-GRB search window, where the frequency n is
    determined by the configuration variable
    :obj:`~gwcelery.conf.joint_mdc_freq` or
    :obj:`~gwcelery.conf.joint_O3_replay_freq`.

    For O3 replay testing with RAVEN pipeline, only run on gracedb-playground.
    """
    if alert['alert_type'] != 'new':
        return
    se_search = alert['object']['preferred_event_data']['search']
    group = alert['object']['preferred_event_data']['group']
    is_gracedb_playground = app.conf['gracedb_host'] \
        == 'gracedb-playground.ligo.org'
    joint_mdc_alert = se_search == 'MDC' and _is_joint_mdc(alert['uid'], 'MDC')
    joint_allsky_alert = se_search == 'AllSky' and \
        _is_joint_mdc(alert['uid'], 'AllSky') and is_gracedb_playground and \
        group in {'CBC', 'Burst'}
    if not (joint_mdc_alert or joint_allsky_alert):
        return

    # Potentially upload 1, 2, or 3 GRB events
    num = 1 + np.random.choice(np.arange(3), p=[.6, .3, .1])

    if joint_mdc_alert:
        # If joint MDC alert, make external event MDC
        if ext_search is None:
            ext_search = 'MDC'
        elif ext_search == 'MDC':
            pass
        else:
            raise ValueError('External search must be "MDC" if MDC superevent')
    # If O3 replay, choose search from acceptable list
    elif ext_search is None:
        # Determine search for external event and then pipelines
        ext_search = \
            (np.random.choice(['GRB', 'SubGRB', 'SubGRBTargeted'],
                              p=[.6, .1, .3])
             if joint_allsky_alert else 'GRB')
    # Choose pipeline(s) based on search
    if ext_search in {'GRB', 'MDC'}:
        pipelines = np.random.choice(['Fermi', 'Swift', 'INTEGRAL', 'AGILE'],
                                     p=[.5, .3, .1, .1],
                                     size=num, replace=False)
    elif ext_search == 'SubGRB':
        pipelines = np.full(num, 'Fermi')
    elif ext_search == 'SubGRBTargeted':
        # Only two current pipelines in targeted search so map 3 => 2
        num = min(num, 2)
        pipelines = np.random.choice(['Fermi', 'Swift'],
                                     p=[.5, .5], size=num, replace=False)
    events = []
    for pipeline in pipelines:
        gpstime = float(alert['object']['t_0'])
        new_time = _offset_time(gpstime, group, pipeline, ext_search)

        # Choose external grb pipeline to simulate
        ext_event = create_upload_external_event(
                        new_time, pipeline, ext_search)

        events.append(ext_event)

    return events, pipelines


@app.task(ignore_result=True,
          shared=False)
def upload_snews_event():
    """Create and upload a SNEWS-like MDC external event."""
    current_time = Time(Time.now(), format='gps').value
    ext_event = create_upload_external_event(current_time, 'SNEWS', 'MDC')
    return ext_event



"""Application configuration for ``gracedb-playground.ligo.org``."""

from base64 import b64encode

from . import *  # noqa: F401, F403

sentry_environment = 'playground'
"""Record this `environment tag
<https://docs.sentry.io/enriching-error-data/environments/>`_ in Sentry log
messages."""

mock_events_simulate_multiple_uploads = True
"""If True, then upload each mock event several times in rapid succession with
random jitter in order to simulate multiple pipeline uploads."""

kafka_alert_config = {
    'scimma': {'url': 'kafka://kafka.scimma.org/igwn.gwalert-playground',
               'suffix': 'avro', 'skymap_encoder': lambda _: _},
    'gcn': {'url': 'kafka://kafka.test.gcn.nasa.gov/igwn.gwalert',
            'suffix': 'json', 'skymap_encoder': lambda b:
            b64encode(b).decode('utf-8')}
}
"""Kafka broker configuration details"""

voevent_broadcaster_address = ':5341'
"""The VOEvent broker will bind to this address to send GCNs.
This should be a string of the form `host:port`. If `host` is empty,
then listen on all available interfaces."""

voevent_broadcaster_whitelist = ['capella2.gsfc.nasa.gov']
"""List of hosts from which the broker will accept connections.
If empty, then completely disable the broker's broadcast capability."""

voevent_receiver_address = '50.116.49.68:8094'
"""The VOEvent listener will connect to this address to receive GCNs. For
options, see `GCN's list of available VOEvent servers
<https://gcn.gsfc.nasa.gov/voevent.html#tc2>`_. If this is an empty string,
then completely disable the GCN listener."""

idq_ok_channels = ['H1:IDQ-OK_OVL_16_4096',
                   'L1:IDQ-OK_OVL_16_4096']
"""Low-latency iDQ OK channel names for O3 replay."""

idq_channels = ['H1:IDQ-FAP_OVL_16_4096',
                'L1:IDQ-FAP_OVL_16_4096']
"""Low-latency iDQ false alarm probability channel names for O3 replay."""



import gzip
import io
from unittest.mock import patch
import json

from astropy.table import Table
import numpy as np
import scipy.stats as stats
import pytest

from ..tasks import gwskynet


def get_toy_3d_fits_filecontents():
    np.random.seed(1000)
    bytesio = io.BytesIO()
    mean = np.random.uniform(low=-1.0, high=1.0)
    sigma = np.random.uniform(low=0.0, high=0.5)
    size = 786432
    prob_cor = np.linspace(mean - 3 * sigma, mean + 3 * sigma, size)
    prob = stats.norm.pdf(prob_cor, mean, sigma)
    table = Table(
        [prob] * 4,
        names=['PROB', 'DISTMU', 'DISTSIGMA', 'DISTNORM'])
    table.meta['LOGBSN'] = 8.0
    table.meta['LOGBCI'] = 3.5
    table.meta['ORDERING'] = 'NESTED'
    table.meta['creator'] = 'BAYESTAR'
    table.meta['distmean'] = np.random.uniform(low=0.0, high=1000)
    table.meta['diststd'] = np.random.uniform(low=0.0, high=250)
    table.meta['instruments'] = ['H1', 'V1']
    with gzip.GzipFile(fileobj=bytesio, mode='wb') as f:
        table.write(f, format='fits')
    return bytesio.getvalue()


def test_gwskynet_annotation():
    outputs = json.loads(gwskynet.gwskynet_annotation(
        get_toy_3d_fits_filecontents()))
    expected_output = {"class_score": 0,
                       "FAP": 1,
                       "FNP": 0}
    for k, v in expected_output.items():
        assert outputs[k] == pytest.approx(v, abs=1e-3)


@patch('gwcelery.tasks.gracedb.download.run',
       return_value=get_toy_3d_fits_filecontents())
@patch('gwcelery.tasks.gracedb.upload.run')
def test_handle_cbc_superevent(mock_upload, mock_download):
    alert = {
        "data": {
            "file_version": 0,
            "name": "GCN_PRELIM_SENT"
        },
        "uid": "TS12345",
        "alert_type": "label_added",
        "object": {
            "preferred_event_data": {
                "group": "CBC",
                "extra_attributes": {
                    'SingleInspiral': [
                        {"ifo": "H1",
                         "snr": 5.6},
                        {"ifo": "L1",
                         "snr": 6.2},
                        {"ifo": "V1",
                         "snr": 3.5}]
                },
                "far": 1e-9,
                "search": 'AllSky',
                "labels": [],
                "offline": False
            }
        }
    }

    gwskynet.handle_cbc_superevent(alert)
    mock_download.assert_called_once_with('bayestar.multiorder.fits',
                                          'TS12345')
    mock_upload.assert_called_once()


@patch('gwcelery.tasks.gracedb.download.run',
       return_value=get_toy_3d_fits_filecontents())
@patch('gwcelery.tasks.gracedb.upload.run')
def test_handle_extreme_high_FAR_cbc_superevent(mock_upload, mock_download):
    alert = {
        "data": {
            "file_version": 0,
            "name": "SKYMAP_READY"
        },
        "uid": "TS12345",
        "alert_type": "label_added",
        "object": {
            "preferred_event_data": {
                "group": "CBC",
                "extra_attributes": {
                    'SingleInspiral': [
                        {"ifo": "H1",
                         "snr": 6.6},
                        {"ifo": "L1",
                         "snr": 6.2},
                        {"ifo": "V1",
                         "snr": 3.5}]
                },
                "far": 1e-04,
                "search": 'AllSky',
                "labels": [],
                "offline": False
            }
        }
    }

    gwskynet.handle_cbc_superevent(alert)
    mock_download.assert_not_called()
    mock_upload.assert_not_called()


@patch('gwcelery.tasks.gracedb.download.run',
       return_value=get_toy_3d_fits_filecontents())
@patch('gwcelery.tasks.gracedb.upload.run')
def test_handle_high_FAR_cbc_superevent(mock_upload, mock_download):
    alert = {
        "data": {
            "file_version": 0,
            "name": "SKYMAP_READY"
        },
        "uid": "TS12345",
        "alert_type": "label_added",
        "object": {
            "preferred_event_data": {
                "group": "CBC",
                "extra_attributes": {
                    'SingleInspiral': [
                        {"ifo": "H1",
                         "snr": 4.6},
                        {"ifo": "L1",
                         "snr": 6.2},
                        {"ifo": "V1",
                         "snr": 4.5}]
                },
                "far": 3e-05,
                "search": 'AllSky',
                "labels": [],
                "offline": False
            }
        }
    }

    gwskynet.handle_cbc_superevent(alert)
    mock_download.assert_called_once_with('bayestar.multiorder.fits',
                                          'TS12345')
    mock_upload.assert_called_once()


@patch('gwcelery.tasks.gracedb.download.run',
       return_value=get_toy_3d_fits_filecontents())
@patch('gwcelery.tasks.gracedb.upload.run')
def test_handle_prefer_update_cbc_superevent(mock_upload, mock_download):
    alert = {
        "data": {
            "file_version": 0,
            "comment": ("Updated superevent parameters: "
                        "preferred_event: TG12344 -> TG12345")
        },
        "uid": "TS12345",
        "alert_type": "log",
        "object": {
            "labels": ["LOW_SIGNIF_PRELIM_SENT"],
            "preferred_event_data": {
                "group": "CBC",
                "extra_attributes": {
                    'SingleInspiral': [
                        {"ifo": "H1",
                         "snr": 4.6},
                        {"ifo": "L1",
                         "snr": 6.2},
                        {"ifo": "V1",
                         "snr": 4.5}]
                },
                "far": 4e-06,
                "search": 'AllSky',
                "labels": [],
                "offline": False
            }
        }
    }

    gwskynet.handle_cbc_superevent(alert)
    mock_download.assert_called_once_with('bayestar.multiorder.fits',
                                          'TS12345')
    mock_upload.assert_called_once()


@patch('gwcelery.tasks.gracedb.download.run',
       return_value=get_toy_3d_fits_filecontents())
@patch('gwcelery.tasks.gracedb.upload.run')
def test_handle_man_prefer_update_cbc_superevent(mock_upload, mock_download):
    alert = {
        "data": {
            "file_version": 0,
            "comment": ("User Albert.Einstein queued a preferred event"
                        " change to TS12345.")
        },
        "uid": "TS12345",
        "alert_type": "log",
        "object": {
            "labels": ["LOW_SIGNIF_PRELIM_SENT"],
            "preferred_event_data": {
                "group": "CBC",
                "extra_attributes": {
                    'SingleInspiral': [
                        {"ifo": "H1",
                         "snr": 4.6},
                        {"ifo": "L1",
                         "snr": 6.2},
                        {"ifo": "V1",
                         "snr": 4.5}]
                },
                "far": 4e-06,
                "search": 'AllSky',
                "labels": [],
                "offline": False
            }
        }
    }

    gwskynet.handle_cbc_superevent(alert)
    mock_download.assert_called_once_with('bayestar.multiorder.fits',
                                          'TS12345')
    mock_upload.assert_called_once()


@patch('gwcelery.tasks.gracedb.download.run',
       return_value=get_toy_3d_fits_filecontents())
@patch('gwcelery.tasks.gracedb.upload.run')
def test_handle_not_prefer_update_cbc_superevent(mock_upload, mock_download):
    alert = {
        "data": {
            "file_version": 0,
            "name": "SKYMAP_READY",
            "labels": ["LOW_SIGNIF_PRELIM_SENT"]
        },
        "uid": "TS12345",
        "alert_type": "update",
        "object": {
            "preferred_event_data": {
                "group": "CBC",
                "extra_attributes": {
                    'SingleInspiral': [
                        {"ifo": "H1",
                         "snr": 4.6},
                        {"ifo": "L1",
                         "snr": 6.2},
                        {"ifo": "V1",
                         "snr": 4.5}]
                },
                "far": 4e-06,
                "search": 'AllSky',
                "labels": [],
                "offline": False
            }
        }
    }

    gwskynet.handle_cbc_superevent(alert)
    mock_download.assert_not_called()
    mock_upload.assert_not_called()


@patch('gwcelery.tasks.gracedb.download.run',
       return_value=get_toy_3d_fits_filecontents())
@patch('gwcelery.tasks.gracedb.upload.run')
def test_handle_earlywarning_cbc_event(mock_upload, mock_download):
    alert = {
        "data": {
            "file_version": 0,
            "name": "SKYMAP_READY"
        },
        "uid": "TS12345",
        "alert_type": "label_added",
        "object": {
            "preferred_event_data": {
                "group": "CBC",
                "extra_attributes": {
                    'SingleInspiral': [
                        {"ifo": "H1",
                         "snr": 6.6},
                        {"ifo": "L1",
                         "snr": 6.2},
                        {"ifo": "V1",
                         "snr": 3.5}]
                },
                "far": 1e-15,
                "search": 'EarlyWarning',
                "labels": [],
                "offline": False
            }
        }
    }

    gwskynet.handle_cbc_superevent(alert)
    mock_download.assert_not_called()
    mock_upload.assert_not_called()



"""GWSkyNet annotation with GWSkyNet model"""
from functools import cache
import json
import re

import numpy as np

from . import gracedb, igwn_alert, superevents
from ..import app
from ..util.tempfile import NamedTemporaryFile

manual_pref_event_change_regexp = re.compile(
    app.conf['views_manual_preferred_event_log_message'].replace('.', '\\.')
    .replace('{}', '.+')
)


@cache
def GWSkyNet_model():
    # FIXME Remove import from function scope once importing GWSkyNet is not a
    # slow operation
    from GWSkyNet import GWSkyNet

    return GWSkyNet.load_GWSkyNet_model()


@app.task(queue='skynet', shared=False)
def gwskynet_annotation(filecontents):
    """Perform the series of tasks necessary for GWSkyNet to

    Parameters
    ----------
    filecontents : bytes
            the sky map downloaded from gracedb
    GWSkyNet_model : keras.engine.functional.Functional object
            the GWSkyNet model used to annotate the events
    """
    # FIXME Remove import from function scope once importing GWSkyNet is not a
    # slow operation
    from GWSkyNet import GWSkyNet

    with NamedTemporaryFile(content=filecontents) as fitsfile:
        GWSkyNet_input = GWSkyNet.prepare_data(fitsfile.name)
    class_score = GWSkyNet.predict(GWSkyNet_model(), GWSkyNet_input)
    FAP, FNP = GWSkyNet.get_rates(class_score)
    fap = FAP[0]
    fnp = FNP[0]
    cs = class_score[0]
    gwskynet_output = {'class_score': cs, 'FAP': fap, 'FNP': fnp}
    return json.dumps(gwskynet_output)


def get_cbc_event_snr(event):
    """Get detector SNRs from the LVAlert packet.

    Parameters
    ----------
    event : dict
        Event dictionary (e.g., the return value from
        :meth:`gwcelery.tasks.gracedb.get_event`, or
        ``preferred_event_data`` in igwn-alert packet.)

    Returns
    -------
    snr : numpy array of floats
        detector SNRs.

    """
    snr = []
    attribs = event['extra_attributes']['SingleInspiral']
    for det in attribs:
        snr.append(det['snr'])
    snr = np.array(snr)
    return snr


@gracedb.task(shared=False)
def _unpack_gwskynet_annotation_and_upload(gwskynet_output, skymap_filename,
                                           graceid):
    filename = 'gwskynet.json'
    gwskynet_output_dict = json.loads(gwskynet_output)
    message = ('GWSkyNet annotation from <a href='
               '"/api/events/{graceid}/files/'
               '{skymap_filename}">'
               '{skymap_filename}</a>.'
               ' GWSkyNet class score: {cs},'
               ' FAP: {FAP}, FNP: {FNP}.').format(
                   graceid=graceid,
                   skymap_filename=skymap_filename,
                   cs=np.round(gwskynet_output_dict['class_score'], 3),
                   FAP=np.round(gwskynet_output_dict['FAP'], 3),
                   FNP=np.round(gwskynet_output_dict['FNP'], 3)
                )
    return gracedb.upload(gwskynet_output, filename, graceid, message=message,
                          tags=['em_follow'])


def _should_annotate(preferred_event, new_label, new_log_comment, labels,
                     alert_type):
    # First check if the event passes all of GWSkyNet's annotation criteria
    SNRs = get_cbc_event_snr(preferred_event)

    if not (preferred_event['search'].lower() == 'allsky' and
            preferred_event['far'] <= app.conf['gwskynet_upper_far_threshold']
            and (SNRs >= app.conf['gwskynet_snr_threshold']).sum() >= 2 and
            np.sqrt(sum(SNRs**2)) >=
            app.conf['gwskynet_network_snr_threshold']):
        return False

    annotate = False
    # Check if the GWSkyNet should annotate in response to this IGWN-Alert
    if alert_type == 'label_added':
        if superevents.should_publish(preferred_event, significant=False) is \
                False and new_label == 'SKYMAP_READY':
            # if the superevent is with FAR higher than the preliminary alert
            # threshold, GWSkyNet will anotate the superevent directly.
            annotate = True
        elif new_label == 'GCN_PRELIM_SENT' or \
                new_label == 'LOW_SIGNIF_PRELIM_SENT':
            # if the FAR is lower than the preliminary alert threshold then
            # GWSkyNet annotates the superevent if the preliminary alert has
            # been sent.
            annotate = True
    elif 'GCN_PRELIM_SENT' not in labels and 'LOW_SIGNIF_PRELIM_SENT' not in \
            labels:
        # GWSkyNet annotations not applied until after initial prelim sent when
        # FAR passes alert threshold
        pass
    elif new_log_comment.startswith('Updated superevent parameters: '
                                    'preferred_event: '):
        # GWSkyNet will also annotate the superevent if the preferred event has
        # changed.
        annotate = True
    elif manual_pref_event_change_regexp.match(new_log_comment):
        # Need to check for a different log comment if the preferred event has
        # been changed manually
        annotate = True

    return annotate


@igwn_alert.handler('superevent',
                    shared=False)
def handle_cbc_superevent(alert):
    """"Annotate the CBC preferred events of superevents using GWSkyNet
    """
    if alert['object']['preferred_event_data']['group'] != 'CBC':
        return

    if alert['alert_type'] != 'label_added' and \
            alert['alert_type'] != 'log':
        return

    superevent_id = alert['uid']
    skymap_filename = 'bayestar.multiorder.fits'
    preferred_event = alert['object']['preferred_event_data']
    new_label = alert['data'].get('name', '')
    new_log_comment = alert['data'].get('comment', '')
    labels = alert['object'].get('labels', [])

    if _should_annotate(preferred_event, new_label, new_log_comment, labels,
                        alert['alert_type']):
        (
            gracedb.download.s(skymap_filename,
                               superevent_id)
            |
            gwskynet_annotation.s()
            |
            _unpack_gwskynet_annotation_and_upload.s(
                skymap_filename, superevent_id)
        ).apply_async()
